import hashlib
import struct
import warnings
import zlib
from functools import wraps

from .. import exception, helpers, primitives
from ..constants import LITTLE_ENDIAN


def _may_recurse(f):
    @wraps(f)
    def safe_recurse(self, *args, **kwargs):
        self._recursion_flag = True
        result = f(self, *args, **kwargs)
        self._recursion_flag = False
        return result

    return safe_recurse


class Checksum(primitives.BasePrimitive):
    """Checksum bound to the block with the specified name.

    The algorithm may be chosen by name with the algorithm parameter, or a custom function may be specified with
    the algorithm parameter.

    The length field is only necessary for custom algorithms.
    When using your own custom checksum function, the return value should be the calculated checksum of the data.

    Function signature:  <function_name>(data_bytes). Returns a number represented as a bytes type.

    Recursive checksums are supported; the checksum field itself will render as all zeros for the sake of checksum
    or length calculations.

    :type  name: str, optional
    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type  block_name: str
    :param block_name: Name of target block for checksum calculations.
    :type  request: boofuzz.Request, optional
    :param request: Request this block belongs to
    :type  algorithm: str, function def name, optional
    :param algorithm: Checksum algorithm to use from this list, default is crc32 (crc32, crc32c, adler32, md5, sha1,
        ipv4, udp). See above for custom checksum function example.

    :type  length: int, optional
    :param length: Length of checksum, auto-calculated by default. Must be specified manually when using custom
        algorithm, defaults to 0
    :type  endian: chr, optional
    :param endian: Endianness of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >), defaults to LITTLE_ENDIAN
    :type  ipv4_src_block_name: str, optional
    :param ipv4_src_block_name: Required for 'udp' algorithm. Name of block yielding IPv4 source address,
        defaults to None
    :type  ipv4_dst_block_name: str, optional
    :param ipv4_dst_block_name: Required for 'udp' algorithm. Name of block yielding IPv4 destination address,
        defaults to None
    :type  fuzzable: bool, optional
    :param fuzzable: Enable/disable fuzzing of this block, defaults to true
    """

    checksum_lengths = {"crc32": 4, "crc32c": 4, "adler32": 4, "md5": 16, "sha1": 20, "ipv4": 2, "udp": 2}

    def __init__(
        self,
        name=None,
        block_name=None,
        request=None,
        algorithm="crc32",
        length=0,
        endian=LITTLE_ENDIAN,
        ipv4_src_block_name=None,
        ipv4_dst_block_name=None,
        *args,
        **kwargs
    ):
        super(Checksum, self).__init__(name=name, *args, **kwargs)

        self._block_name = block_name
        self._request = request
        self._algorithm = algorithm
        self._length = length
        self._endian = endian
        self._ipv4_src_block_name = ipv4_src_block_name
        self._ipv4_dst_block_name = ipv4_dst_block_name

        if not self._length and self._algorithm in self.checksum_lengths:
            self._length = self.checksum_lengths[self._algorithm]

        # Edge cases and a couple arbitrary strings (all 1s, all Es)
        self._fuzz_library = [
            b"\x00" * self._length,
            b"\x11" * self._length,
            b"\xEE" * self._length,
            b"\xFF" * self._length,
            b"\xFF" * (self._length - 1) + b"\xFE",
            b"\x00" * (self._length - 1) + b"\x01",
        ]

        if self._algorithm == "udp":
            if not self._ipv4_src_block_name:
                raise exception.SullyRuntimeError("'udp' checksum algorithm requires ipv4_src_block_name")
            if not self._ipv4_dst_block_name:
                raise exception.SullyRuntimeError("'udp' checksum algorithm requires ipv4_dst_block_name")

        self._rendered = self._get_dummy_value()

        # Set the recursion flag before calling a method that may cause a recursive loop.
        self._recursion_flag = False

    def encode(self, value, mutation_context):
        if value is None:
            if self._recursion_flag or self._request is None:
                self._rendered = self._get_dummy_value()
            else:
                self._rendered = self._checksum(
                    data=self._render_block(self._block_name, mutation_context=mutation_context),
                    ipv4_src=self._render_block(self._ipv4_src_block_name, mutation_context=mutation_context),
                    ipv4_dst=self._render_block(self._ipv4_dst_block_name, mutation_context=mutation_context),
                )
            return helpers.str_to_bytes(self._rendered)
        else:
            return value

    def _get_dummy_value(self):
        return self._length * "\x00"

    @_may_recurse
    def _render_block(self, block_name, mutation_context):
        return (
            self._request.resolve_name(self.context_path, block_name).render(mutation_context=mutation_context)
            if block_name is not None and self._request is not None
            else None
        )

    def _checksum(self, data, ipv4_src, ipv4_dst):
        """
        Calculate and return the checksum (in raw bytes) of data.

        :param data Data on which to calculate checksum.
        :type data bytes

        :rtype:  bytes
        :return: Checksum.
        """
        if isinstance(self._algorithm, str):
            if self._algorithm == "crc32":
                check = struct.pack(self._endian + "L", (zlib.crc32(data) & 0xFFFFFFFF))

            elif self._algorithm == "crc32c":
                try:
                    import crc32c  # pytype: disable=import-error
                except ImportError:
                    warnings.warn(
                        "Importing crc32c package failed. Please install it using pip.", UserWarning, stacklevel=2
                    )
                    raise
                check = struct.pack(self._endian + "L", crc32c.crc32(data))

            elif self._algorithm == "adler32":
                check = struct.pack(self._endian + "L", (zlib.adler32(data) & 0xFFFFFFFF))

            elif self._algorithm == "ipv4":
                check = struct.pack(self._endian + "H", helpers.ipv4_checksum(data))

            elif self._algorithm == "udp":
                return struct.pack(
                    self._endian + "H", helpers.udp_checksum(msg=data, src_addr=ipv4_src, dst_addr=ipv4_dst)
                )

            elif self._algorithm == "md5":
                digest = hashlib.md5(data).digest()

                # TODO: is this right?
                if self._endian == ">":
                    (a, b, c, d) = struct.unpack("<LLLL", digest)
                    digest = struct.pack(">LLLL", a, b, c, d)

                check = digest

            elif self._algorithm == "sha1":
                digest = hashlib.sha1(data).digest()

                # TODO: is this right?
                if self._endian == ">":
                    (a, b, c, d, e) = struct.unpack("<LLLLL", digest)
                    digest = struct.pack(">LLLLL", a, b, c, d, e)

                check = digest

            else:
                raise exception.SullyRuntimeError("INVALID CHECKSUM ALGORITHM SPECIFIED: %s" % self._algorithm)
        else:
            check = self._algorithm(data)

        if self._length:
            return check[: self._length]
        else:
            return check

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self._name)

    def __len__(self):
        return self._length



import abc


# abc.ABCMeta is the metaclass in both python 2 and 3
class ITargetConnection(metaclass=abc.ABCMeta):
    """
    Interface for connections to fuzzing targets.
    Target connections may be opened and closed multiple times. You must open before using send/recv and close
    afterwards.

    .. versionchanged:: 0.2.0
        ITargetConnection has been moved into the connections subpackage.
        The full path is now boofuzz.connections.itarget_connection.ITargetConnection
    """

    @abc.abstractmethod
    def close(self):
        """
        Close connection.

        :return: None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def open(self):
        """
        Opens connection to the target. Make sure to call close!

        :return: None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def recv(self, max_bytes):
        """
        Receive up to max_bytes data.

        :param max_bytes: Maximum number of bytes to receive.
        :type max_bytes: int

        :return: Received data. bytes('') if no data is received.
        :rtype: bytes
        """
        raise NotImplementedError

    @abc.abstractmethod
    def send(self, data):
        """
        Send data to the target.

        :param data: Data to send.

        :return: Number of bytes actually sent.
        :rtype: int
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def info(self):
        """Return description of connection info.

        E.g., "127.0.0.1:2121"

        Returns:
            str: Connection info descrption
        """
        raise NotImplementedError



# Misc Types
from .. import blocks, exception, helpers, primitives


class DNSHostname(blocks.Block):
    def __init__(self, name, request, value, options=None):
        if not options:
            options = {}

        super(DNSHostname).__init__(name, request)

        self.value = value
        self.options = options

        if not self.value:
            raise exception.SullyRuntimeError("MISSING LEGO.tag DEFAULT VALUE")

        self.push(primitives.String())

    def render(self):
        """
        We overload and extend the render routine in order to properly insert substring lengths.
        """

        # let the parent do the initial render.
        blocks.Block.render(self)

        new_str = b""

        # replace dots (.) with the substring length.
        for part in self._rendered.split(b"."):
            new_str += bytes(len(part)) + part

        # be sure to null terminate too.
        self._rendered = new_str + b"\x00"

        return helpers.str_to_bytes(self._rendered)


class Tag(blocks.Block):
    def __init__(self, name, request, value, options=None):
        if not options:
            options = {}

        super(Tag, self).__init__(name, request)

        self.value = value
        self.options = options

        if not self.value:
            raise exception.SullyRuntimeError("MISSING LEGO.tag DEFAULT VALUE")

        # <example>
        # [delim][string][delim]

        self.push(primitives.Delim(name=name + "_DELIM1", default_value="<"))
        self.push(primitives.String(name=name + "_STR", default_value=""))
        self.push(primitives.Delim(name=name + "_DELIM2", default_value=">"))



import struct

from .. import helpers
from ..constants import LITTLE_ENDIAN
from ..fuzzable import Fuzzable


def binary_string_to_int(binary):
    """
    Convert a binary string to a decimal number.

    @type  binary: str
    @param binary: Binary string

    @rtype:  int
    @return: Converted bit string
    """

    return int(binary, 2)


def int_to_binary_string(number, bit_width):
    """
    Convert a number to a binary string.

    @type  number:    int
    @param number:    (Optional, def=self._value) Number to convert
    @type  bit_width: int
    @param bit_width: (Optional, def=self.width) Width of bit string

    @rtype:  str
    @return: Bit string
    """
    return "".join(map(lambda x: str((number >> x) & 1), range(bit_width - 1, -1, -1)))


class BitField(Fuzzable):
    """
    The bit field primitive represents a number of variable length and is used to define all other integer types.

    :type  name: str, optional
    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type  default_value: int, optional
    :param default_value: Default integer value, defaults to 0
    :type  width: int, optional
    :param width: Width in bits, defaults to 8
    :type  max_num: int, optional
    :param max_num: Maximum number to iterate up to, defaults to None
    :type  endian: char, optional
    :param endian: Endianness of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >), defaults to LITTLE_ENDIAN
    :type  output_format: str, optional
    :param output_format: Output format, "binary" or "ascii", defaults to binary
    :type  signed: bool, optional
    :param signed: Make size signed vs. unsigned (applicable only with format="ascii"), defaults to False
    :type  full_range: bool, optional
    :param full_range: If enabled the field mutates through *all* possible values, defaults to False
    :type  fuzz_values: list, optional
    :param fuzz_values: List of custom fuzz values to add to the normal mutations, defaults to None
    :type  fuzzable: bool, optional
    :param fuzzable: Enable/disable fuzzing of this primitive, defaults to true
    """

    def __init__(
        self,
        name=None,
        default_value=0,
        width=8,
        max_num=None,
        endian=LITTLE_ENDIAN,
        output_format="binary",
        signed=False,
        full_range=False,
        *args,
        **kwargs
    ):
        super(BitField, self).__init__(name=name, default_value=default_value, *args, **kwargs)

        assert isinstance(width, int), "width must be an integer!"

        self.width = width
        self.max_num = max_num
        self.endian = endian
        self.format = output_format
        self.signed = signed
        self.full_range = full_range

        if not self.max_num:
            self.max_num = binary_string_to_int("1" + "0" * width)

        assert isinstance(self.max_num, int), "max_num must be an integer!"

        if not self.full_range:
            # try only "smart" values.
            interesting_boundaries = [
                0,
                self.max_num // 2,
                self.max_num // 3,
                self.max_num // 4,
                self.max_num // 8,
                self.max_num // 16,
                self.max_num // 32,
                self.max_num,
            ]

            # Contract: sort in ascending order required for deduplication.
            interesting_boundaries.sort()

            self._interesting_boundaries = interesting_boundaries
        else:
            self._interesting_boundaries = []

    def _iterate_fuzz_lib(self):
        if self.full_range:
            for i in range(0, self.max_num):
                yield i
        else:
            # To avoid duplication of mutatation values, we introduce gradually
            # increasing lower border. Combined with interesting boundaries
            # sorted in ascending order, it's possible to avoid duplicates.
            # Deduplication could also be done with intermediate 'set' construction,
            # but it might be costly or impossible if the number of values is huge.
            # We don't expect negative bit field values with Python integers,
            # the 0-value for mutation is possible with the given starting 'lower_border'.
            lower_border = -1
            for boundary in self._interesting_boundaries:
                for v in self._yield_integer_boundaries(boundary, lower_border):
                    lower_border = v
                    yield v
        # TODO Add a way to inject a list of fuzz values
        # elif isinstance(default_value, (list, tuple)):
        # for val in iter(default_value):
        #    yield val

        # TODO: Add injectable arbitrary bit fields

    def _yield_integer_boundaries(self, integer, lower_border):
        """
        Add the supplied integer and border cases to the integer fuzz heuristics library.

        @type  integer: int
        @param integer: int to append to fuzz heuristics
        @type  lower_border: int
        @param lower_border: int bottom limit for border cases, so all values must be strictly greater
        """
        # Contract: generate values in ascending order, otherwise deduplication logics might break.
        for i in range(-10, 10):
            case = integer + i
            if lower_border < case < self.max_num:
                # some day: if case not in self._user_provided_values
                yield case

    def encode(self, value, mutation_context):
        temp = self._render_int(
            value, output_format=self.format, bit_width=self.width, endian=self.endian, signed=self.signed
        )
        return helpers.str_to_bytes(temp)

    def mutations(self, default_value):
        for val in self._iterate_fuzz_lib():
            yield val

    @staticmethod
    def _render_int(value, output_format, bit_width, endian, signed):
        """
        Convert value to a bit or byte string.

        Args:
            value (int): Value to convert to a byte string.
            output_format (str): "binary" or "ascii"
            bit_width (int): Width of output in bits.
            endian: BIG_ENDIAN or LITTLE_ENDIAN
            signed (bool):

        Returns:
            str: value converted to a byte string
        """
        if output_format == "binary":
            bit_stream = ""
            rendered = b""

            # pad the bit stream to the next byte boundary.
            if bit_width % 8 == 0:
                bit_stream += int_to_binary_string(value, bit_width)
            else:
                bit_stream = "0" * (8 - (bit_width % 8))
                bit_stream += int_to_binary_string(value, bit_width)

            # convert the bit stream from a string of bits into raw bytes.
            for i in range(len(bit_stream) // 8):
                chunk_min = 8 * i
                chunk_max = chunk_min + 8
                chunk = bit_stream[chunk_min:chunk_max]
                rendered += struct.pack("B", binary_string_to_int(chunk))

            # if necessary, convert the endianness of the raw bytes.
            if endian == LITTLE_ENDIAN:
                # reverse the bytes
                rendered = rendered[::-1]

            _rendered = rendered
        else:
            # Otherwise we have ascii/something else
            # if the sign flag is raised and we are dealing with a signed integer (first bit is 1).
            if signed and int_to_binary_string(value, bit_width)[0] == "1":
                max_num = binary_string_to_int("1" + "0" * (bit_width - 1))
                # chop off the sign bit.
                val = value & binary_string_to_int("1" * (bit_width - 1))

                # account for the fact that the negative scale works backwards.
                val = max_num - val - 1

                # toss in the negative sign.
                _rendered = "%d" % ~val

            # unsigned integer or positive signed integer.
            else:
                _rendered = "%d" % value
        return _rendered



"""Demo of a very simple protocol definition using the float primitive."""
from boofuzz import *


def main():
    session = Session(target=Target(connection=UDPSocketConnection("127.0.0.1", 9999)))
    define_floats(session)
    session.fuzz()


def define_floats(session):
    s_initialize("float-test")

    # Creating 10 float x with 5.0 <= x.x <= 10.0 and with random seed = "first_seed"
    s_float(10.0, f_min=5.0, f_max=15.0, max_mutations=10, seed="fist_seed", fuzzable=True, name="first_float")
    s_static(",")

    # Creating 10 floats x with 10.0 <= x.xx <= 20.0 adn with random seed = "second_seed"
    s_float(
        12.34,
        s_format=".2f",
        f_min=10.0,
        f_max=20.0,
        max_mutations=10,
        seed="second_seed",
        fuzzable=True,
        name="second_float",
    )
    s_static(",")

    # Creating 25 floats x with 20.0 <= x.xxxx <= 40.0, without a random seed
    s_float(25.1234, f_min=20.0, f_max=40.0, max_mutations=25, fuzzable=True, name="third_float")
    s_static(",")

    # Creates a float, that will not be fuzzed. The Value is static.
    s_float(42.0, fuzzable=False, name="static_float_value")
    s_static(",")

    # Creates a float, with following format: xxxx.xx (fills in leading zeros)
    s_float(15.0, s_format="07.2f", f_min=0.0, f_max=100.0, max_mutations=10, fuzzable=True, name="float_leading_zeros")
    s_static(",")

    # Creates a float encoded as IEEE 754 floating point (big endian)
    s_float(-12000.25, max_mutations=10, fuzzable=True, encode_as_ieee_754=True, endian="big", name="float_ieee_754")
    s_static(",")

    # Creates a float encoded as IEEE 754 floating point (little endian)
    s_float(
        -12000.25, max_mutations=10, fuzzable=True, encode_as_ieee_754=True, endian="little", name="ieee_little_endian"
    )
    s_static("\r\n")

    session.connect(s_get("float-test"))


if __name__ == "__main__":
    main()



import warnings

from boofuzz.connections import itarget_connection


class NETCONFConnection(itarget_connection.ITargetConnection):
    """
    ITargetConnection implementation for NETCONF server connections.
    Unlike ITargetConnection, NETCONFConnection works with utf-8 encoded strings
    instead of bytes.

    Args:
        host (str): IP address of NETCONF server.
        port (int): port of NETCONF server.
        username (str): NETCONF server login username.
        password (str): NETCONF server login password.
        datastore (str): NETCONF server datastore to send fuzz requests to.
        hostkey_verify(bool): Verify SSH host key when connecting to NETCONF server
    """

    def __init__(self, host, port, username, password, datastore, hostkey_verify):
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.datastore = datastore
        self.hostkey_verify = hostkey_verify
        self._received_data = None
        self._conn = None

    def open(self):
        try:
            from ncclient import manager  # pytype: disable=import-error
        except ImportError:
            warnings.warn("Importing ncclient package failed. Please install it using pip.", UserWarning)
            raise

        self._conn = manager.connect(
            host=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
            hostkey_verify=self.hostkey_verify,
        )

    def close(self):
        if self._conn.connected:
            self._conn.close_session()

    def recv(self, max_bytes):
        """
        Receive data from the NETCONF server.

        Args:
            max_bytes (int): Maximum number of bytes to receive. Currently ignored.

        Returns:
            str: utf-8 encoded XML response
        """

        data = self._received_data
        self._received_data = None

        if data is None:
            data = ""

        return data

    def send(self, data):
        """
        Send an edit-config request to the NETCONF server.

        Args:
            data (str): XML data for an XML edit_config request. Should be a
            string with utf-8 encoding.
        """

        data = data.decode("utf-8")

        # store data for later recv() calls
        self._received_data = self._conn.edit_config(target=self.datastore, config=data)

    def get_raw_conn(self):
        return self._conn

    @property
    def info(self):
        return (
            "host: {host}, port: {port}, username: {username},"
            " datastore: {datastore}, hostkey_verify: {hostkey_verify}".format(
                host=self.host,
                port=self.port,
                username=self.username,
                datastore=self.datastore,
                hostkey_verify=self.hostkey_verify,
            )
        )



import functools
import sys


from . import blocks, exception, legos, primitives
from .blocks import Aligned, Block, Checksum, Repeat, Request, REQUESTS, Size
from .cli import main_helper
from .connections import (
    BaseSocketConnection,
    FileConnection,
    ip_constants,
    ISerialLike,
    ITargetConnection,
    NETCONFConnection,
    RawL2SocketConnection,
    RawL3SocketConnection,
    SerialConnection,
    SerialConnectionLowLevel,
    SocketConnection,
    SSLSocketConnection,
    TCPSocketConnection,
    UDPSocketConnection,
    UnixSocketConnection,
)
from .constants import BIG_ENDIAN, DEFAULT_PROCMON_PORT, LITTLE_ENDIAN
from .event_hook import EventHook
from .exception import BoofuzzFailure, MustImplementException, SizerNotUtilizedError, SullyRuntimeError
from .fuzz_logger import FuzzLogger
from .fuzz_logger_csv import FuzzLoggerCsv
from .fuzz_logger_curses import FuzzLoggerCurses
from .fuzz_logger_text import FuzzLoggerText
from .fuzzable import Fuzzable
from .fuzzable_block import FuzzableBlock
from .ifuzz_logger import IFuzzLogger
from .ifuzz_logger_backend import IFuzzLoggerBackend
from .monitors import BaseMonitor, CallbackMonitor, NetworkMonitor, pedrpc, ProcessMonitor
from .utils.process_monitor_local import ProcessMonitorLocal
from .primitives import (
    BasePrimitive,
    BitField,
    Byte,
    Bytes,
    Delim,
    DWord,
    Float,
    FromFile,
    Group,
    Mirror,
    QWord,
    RandomData,
    Simple,
    Static,
    String,
    Word,
)
from .repeater import CountRepeater, Repeater, TimeRepeater
from .sessions import open_test_run, Session, Target
from .protocol_session import ProtocolSession
from .protocol_session_reference import ProtocolSessionReference

# workaround to make Tornado work in Python 3.8
# https://github.com/tornadoweb/tornado/issues/2608
if sys.platform == "win32" and sys.version_info >= (3, 8):
    import asyncio

    # noinspection PyUnresolvedReferences
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())  # pytype: disable=module-attr

__all__ = [
    "Aligned",
    "BaseMonitor",
    "BasePrimitive",
    "BaseSocketConnection",
    "BIG_ENDIAN",
    "BitField",
    "Block",
    "blocks",
    "BoofuzzFailure",
    "Byte",
    "Bytes",
    "CallbackMonitor",
    "Checksum",
    "CountRepeater",
    "DEFAULT_PROCMON_PORT",
    "Delim",
    "DWord",
    "EventHook",
    "exception",
    "FileConnection",
    "Float",
    "FromFile",
    "Fuzzable",
    "FuzzableBlock",
    "FuzzLogger",
    "FuzzLoggerCsv",
    "FuzzLoggerCurses",
    "FuzzLoggerText",
    "Group",
    "IFuzzLogger",
    "IFuzzLoggerBackend",
    "ip_constants",
    "ISerialLike",
    "ITargetConnection",
    "NETCONFConnection",
    "legos",
    "LITTLE_ENDIAN",
    "main_helper",
    "Mirror",
    "MustImplementException",
    "NetworkMonitor",
    "open_test_run",
    "pedrpc",
    "primitives",
    "ProcessMonitor",
    "ProcessMonitorLocal",
    "QWord",
    "RandomData",
    "RawL2SocketConnection",
    "RawL3SocketConnection",
    "Repeat",
    "Repeater",
    "Request",
    "REQUESTS",
    "s_aligned",
    "s_bigword",
    "s_binary",
    "s_bit",
    "s_bit_field",
    "s_bits",
    "s_block",
    "s_block_end",
    "s_block_start",
    "s_byte",
    "s_bytes",
    "s_char",
    "s_checksum",
    "s_cstring",
    "s_delim",
    "s_double",
    "s_dunno",
    "s_dword",
    "s_float",
    "s_from_file",
    "s_get",
    "s_group",
    "s_hex_dump",
    "s_initialize",
    "s_int",
    "s_intelword",
    "s_lego",
    "s_long",
    "s_mirror",
    "s_num_mutations",
    "s_qword",
    "s_random",
    "s_raw",
    "s_repeat",
    "s_repeater",
    "s_short",
    "s_size",
    "s_sizer",
    "s_static",
    "s_simple",
    "s_string",
    "s_switch",
    "s_unknown",
    "s_update",
    "s_word",
    "SerialConnection",
    "SerialConnectionLowLevel",
    "Session",
    "Size",
    "SizerNotUtilizedError",
    "SocketConnection",
    "SSLSocketConnection",
    "Simple",
    "Static",
    "String",
    "SullyRuntimeError",
    "Target",
    "TCPSocketConnection",
    "ProtocolSession",
    "ProtocolSessionReference",
    "TimeRepeater",
    "UDPSocketConnection",
    "UnixSocketConnection",
    "Word",
]


# REQUEST MANAGEMENT
def s_get(name=None):
    """
    Return the request with the specified name or the current request if name is not specified. Use this to switch from
    global function style request manipulation to direct object manipulation. Example::

        req = s_get("HTTP BASIC")
        print(req.num_mutations())

    The selected request is also set as the default current. (ie: s_switch(name) is implied).

    :type  name: str
    :param name: (Optional, def=None) Name of request to return or current request if name is None.

    :rtype:  blocks.Request
    :return: The requested request.
    """

    if not name:
        return blocks.CURRENT

    # ensure this gotten request is the new current.
    s_switch(name)

    if name not in blocks.REQUESTS:
        raise exception.SullyRuntimeError("blocks.REQUESTS NOT FOUND: %s" % name)

    return blocks.REQUESTS[name]


def s_initialize(name):
    """
    Initialize a new block request. All blocks / primitives generated after this call apply to the named request.
    Use s_switch() to jump between factories.

    :type  name: str
    :param name: Name of request
    """
    if name in blocks.REQUESTS:
        raise exception.SullyRuntimeError("blocks.REQUESTS ALREADY EXISTS: %s" % name)

    blocks.REQUESTS[name] = Request(name)
    blocks.CURRENT = blocks.REQUESTS[name]


def s_num_mutations():
    """
    Determine the number of repetitions we will be making.

    :rtype:  int
    :return: Number of mutated forms this primitive can take.
    """

    return blocks.CURRENT.get_num_mutations()


def s_switch(name):
    """
    Change the current request to the one specified by "name".

    :type  name: str
    :param name: Name of request
    """

    if name not in blocks.REQUESTS:
        raise exception.SullyRuntimeError("blocks.REQUESTS NOT FOUND: %s" % name)

    blocks.CURRENT = blocks.REQUESTS[name]


# ## BLOCK MANAGEMENT


def s_block(name=None, group=None, encoder=None, dep=None, dep_value=None, dep_values=None, dep_compare="=="):
    """
    Open a new block under the current request. The returned instance supports the "with" interface so it will
    be automatically closed for you::

        with s_block("header"):
            s_static("\\x00\\x01")
            if s_block_start("body"):
                ...

    :type  name:        str, optional
    :param name:        Name of block being opened
    :type  group:       str, optional
    :param group:       (Optional, def=None) Name of group to associate this block with
    :type  encoder:     Function Pointer, optional
    :param encoder:     (Optional, def=None) Optional pointer to a function to pass rendered data to prior to return
    :type  dep:         str, optional
    :param dep:         (Optional, def=None) Optional primitive whose specific value this block is dependant on
    :type  dep_value:   bytes, optional
    :param dep_value:   (Optional, def=None) Value that field "dep" must contain for block to be rendered
    :type  dep_values:  List of bytes, optional
    :param dep_values:  (Optional, def=None) Values that field "dep" may contain for block to be rendered
    :type  dep_compare: str, optional
    :param dep_compare: (Optional, def="==") Comparison method to use on dependency (==, !=, >, >=, <, <=)
    """

    class ScopedBlock:
        def __init__(self, block):
            self.block = block

        def __enter__(self):
            """
            Setup before entering the "with" statement body
            """
            return self.block

        def __exit__(self, type, value, traceback):
            """
            Cleanup after executing the "with" statement body
            """
            # Automagically close the block when exiting the "with" statement
            s_block_end()

    block = s_block_start(
        name,
        request=blocks.CURRENT,
        group=group,
        encoder=encoder,
        dep=dep,
        dep_value=dep_value,
        dep_values=dep_values,
        dep_compare=dep_compare,
    )

    return ScopedBlock(block)


def s_aligned(modulus=1, pattern=b"\x00", name=None):
    """FuzzableBlock that aligns its contents to a certain number of bytes

    :type  modulus:     int, optional
    :param modulus:     Pad length of child content to this many bytes, defaults to 1
    :type  pattern:     bytes, optional
    :param pattern:     Pad using these byte(s)
    :type  name:        str, optional
    :param name:        Name, for referencing later. Names should always be provided, but if not, a default name will
                        be given, defaults to None
    """

    class ScopedAligned:
        def __init__(self, aligned):
            self.aligned = aligned

        def __enter__(self):
            """
            Setup before entering the "with" statement body
            """
            return self.aligned

        def __exit__(self, type, value, traceback):
            """
            Cleanup after executing the "with" statement body
            """
            blocks.CURRENT.pop()

    aligned = Aligned(name=name, modulus=modulus, pattern=pattern, fuzzable=True)
    blocks.CURRENT.push(aligned)

    return ScopedAligned(aligned)


def s_block_start(name=None, *args, **kwargs):
    """
    Open a new block under the current request. This routine always returns an instance so you can make your fuzzer
    pretty with indenting::

        if s_block_start("header"):
            s_static("\\x00\\x01")
            if s_block_start("body"):
                ...
        s_block_close()

    :note Prefer using s_block to this function directly
    :see s_block
    """
    block = Block(name=name, *args, **kwargs)
    blocks.CURRENT.push(block)

    return block


# noinspection PyUnusedLocal
def s_block_end(name=None):
    """
    Close the last opened block. Optionally specify the name of the block being closed (purely for aesthetic purposes).

    :type  name: str
    :param name: (Optional, def=None) Name of block to closed.
    """
    blocks.CURRENT.pop()


def s_checksum(
    block_name=None,
    algorithm="crc32",
    length=0,
    endian=LITTLE_ENDIAN,
    fuzzable=True,
    name=None,
    ipv4_src_block_name=None,
    ipv4_dst_block_name=None,
):
    """
    Checksum bound to the block with the specified name.

    The algorithm may be chosen by name with the algorithm parameter, or a custom function may be specified with
    the algorithm parameter.

    The length field is only necessary for custom algorithms.

    Recursive checksums are supported; the checksum field itself will render as all zeros for the sake of checksum
    or length calculations.

    :type  block_name: str, optional
    :param block_name: Name of target block for checksum calculations.
    :type  algorithm: str, function, optional
    :param algorithm: Checksum algorithm to use. (crc32, crc32c, adler32, md5, sha1, ipv4, udp)
        Pass a function to use a custom algorithm. This function has to take and return byte-type data,
        defaults to crc32
    :type  length: int, optional
    :param length: Length of checksum, auto-calculated by default. Must be specified manually when using custom
        algorithm, defaults to 0
    :type  endian: chr, optional
    :param endian: Endianness of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >), defaults to LITTLE_ENDIAN
    :type  fuzzable:   bool, optional
    :param fuzzable:   Enable/disable fuzzing.
    :type  name: str, optional
    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type  ipv4_src_block_name: str, optional
    :param ipv4_src_block_name: Required for 'udp' algorithm. Name of block yielding IPv4 source address,
        defaults to None
    :type  ipv4_dst_block_name: str, optional
    :param ipv4_dst_block_name: Required for 'udp' algorithm. Name of block yielding IPv4 destination address,
        defaults to None
    """

    # you can't add a checksum for a block currently in the stack.
    if block_name in blocks.CURRENT.block_stack:
        raise exception.SullyRuntimeError("CAN N0T ADD A CHECKSUM FOR A BLOCK CURRENTLY IN THE STACK")

    checksum = Checksum(
        name=name,
        block_name=block_name,
        request=blocks.CURRENT,
        algorithm=algorithm,
        length=length,
        endian=endian,
        fuzzable=fuzzable,
        ipv4_src_block_name=ipv4_src_block_name,
        ipv4_dst_block_name=ipv4_dst_block_name,
    )
    blocks.CURRENT.push(checksum)


def s_repeat(block_name=None, min_reps=0, max_reps=25, step=1, variable=None, fuzzable=True, name=None):
    """
    Repeat the rendered contents of the specified block cycling from min_reps to max_reps counting by step. By
    default renders to nothing. This block modifier is useful for fuzzing overflows in table entries. This block
    modifier MUST come after the block it is being applied to.

    :see: Aliases: s_repeater()

    :type  block_name: str
    :param block_name: (Optional, def=None) Name of block to repeat
    :type  min_reps:   int
    :param min_reps:   (Optional, def=0) Minimum number of block repetitions
    :type  max_reps:   int
    :param max_reps:   (Optional, def=25) Maximum number of block repetitions
    :type  step:       int
    :param step:       (Optional, def=1) Step count between min and max reps
    :type  variable:   Sulley Integer Primitive
    :param variable:   (Optional, def=None) An integer primitive which will specify the number of repitions
    :type  fuzzable:   bool
    :param fuzzable:   (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  name:       str
    :param name:       (Optional, def=None) Specifying a name gives you direct access to a primitive
    """

    blocks.CURRENT.push(
        Repeat(
            name=name,
            block_name=block_name,
            request=blocks.CURRENT,
            min_reps=min_reps,
            max_reps=max_reps,
            step=step,
            variable=variable,
            fuzzable=fuzzable,
        )
    )


def s_size(
    block_name=None,
    offset=0,
    length=4,
    endian=LITTLE_ENDIAN,
    output_format="binary",
    inclusive=False,
    signed=False,
    math=None,
    fuzzable=True,
    name=None,
):
    """
    Create a sizer block bound to the block with the specified name. You *can not* create a sizer for any
    currently open blocks.

    :see: Aliases: s_sizer()

    :type  block_name:    str, optional
    :param block_name:    Name of block to apply sizer to.
    :type  offset:        int, optional
    :param offset:        Offset for calculated size value, defaults to 0
    :type  length:        int, optional
    :param length:        Length of sizer, defaults to 4
    :type  endian:        chr, optional
    :param endian:        Endianness of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >), defaults to LITTLE_ENDIAN
    :type  output_format: str, optional
    :param output_format: Output format, "binary" or "ascii", defaults to binary
    :type  inclusive:     bool, optional
    :param inclusive:     Should the sizer count its own length? Defaults to False
    :type  signed:        bool, optional
    :param signed:        Make size signed vs. unsigned (applicable only with format="ascii"), defaults to False
    :type  math:          def, optional
    :param math:          Apply the mathematical op defined in this function to the size, defaults to None
    :type  fuzzable:      bool
    :param fuzzable:      (Optional, def=True) Enable/disable fuzzing of this sizer
    :type  name:          str
    :param name:          Name of this sizer field
    """

    blocks.CURRENT.push(
        Size(
            name=name,
            block_name=block_name,
            request=blocks.CURRENT,
            offset=offset,
            length=length,
            endian=endian,
            output_format=output_format,
            inclusive=inclusive,
            signed=signed,
            math=math,
            fuzzable=fuzzable,
        )
    )


def s_update(name, value):
    """
    Update the value of the named primitive in the currently open request.

    :type  name:  str
    :param name:  Name of object whose value we wish to update
    :type  value: Mixed
    :param value: Updated value
    """

    if name not in map(lambda o: o.qualified_name, blocks.CURRENT.walk()):
        raise exception.SullyRuntimeError("NO OBJECT WITH NAME '%s' FOUND IN CURRENT REQUEST" % name)

    blocks.CURRENT.names[name]._default_value = value


# PRIMITIVES


def s_binary(value, name=None):
    """
    Parse a variable format binary string into a static value and push it onto the current block stack.

    :type  value: str
    :param value: Variable format binary string
    :type  name:  str
    :param name:  (Optional, def=None) Specifying a name gives you direct access to a primitive
    """

    # parse the binary string into.
    parsed = value
    parsed = parsed.replace(" ", "")
    parsed = parsed.replace("\t", "")
    parsed = parsed.replace("\r", "")
    parsed = parsed.replace("\n", "")
    parsed = parsed.replace(",", "")
    parsed = parsed.replace("0x", "")
    parsed = parsed.replace("\\x", "")

    value = b""
    while parsed:
        pair = parsed[:2]
        parsed = parsed[2:]

        value += int(pair, 16).to_bytes(1, "little")

    blocks.CURRENT.push(Static(name=name, default_value=parsed))


def s_delim(value=" ", fuzzable=True, name=None):
    """
    Push a delimiter onto the current block stack.

    :type  value:    Character
    :param value:    (Optional, def=" ")Original value
    :type  fuzzable: bool
    :param fuzzable: (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  name:     str
    :param name:     (Optional, def=None) Specifying a name gives you direct access to a primitive
    """

    blocks.CURRENT.push(Delim(name=name, default_value=value, fuzzable=fuzzable))


def s_group(name=None, values=None, default_value=None):
    """
    This primitive represents a list of static values, stepping through each one on mutation. You can tie a block
    to a group primitive to specify that the block should cycle through all possible mutations for *each* value
    within the group. The group primitive is useful for example for representing a list of valid opcodes.

    :type  name:            str
    :param name:            (Optional, def=None) Name of group
    :type  values:          List or raw data
    :param values:          (Optional, def=None) List of possible raw values this group can take.
    :type  default_value:   str or bytes
    :param default_value:   (Optional, def=None) Specifying a value when fuzzing() is complete
    """

    blocks.CURRENT.push(Group(name=name, default_value=default_value, values=values))


# noinspection PyCallingNonCallable
def s_lego(lego_type, value=None, options=()):
    """
    Legos are pre-built blocks... TODO: finish this doc

    :type  lego_type:   str
    :param lego_type:   Function that represents a lego

    :param value:       Original value

    :param options:     Options to pass to lego.
    """

    # as legos are blocks they must have a name.
    # generate a unique name for this lego.
    name = "LEGO_%08x" % len(blocks.CURRENT.names)

    if lego_type not in legos.BIN:
        raise exception.SullyRuntimeError("INVALID LEGO TYPE SPECIFIED: %s" % lego_type)
    lego = legos.BIN[lego_type](name, blocks.CURRENT, value, options)

    # push the lego onto the stack and immediately pop to close the block.
    blocks.CURRENT.push(lego)
    blocks.CURRENT.pop()


def s_random(value="", min_length=0, max_length=1, num_mutations=25, fuzzable=True, step=None, name=None):
    """
    Generate a random chunk of data while maintaining a copy of the original. A random length range can be specified.
    For a static length, set min/max length to be the same.

    :type  value:         str or bytes
    :param value:         (Optional, def="") Original value
    :type  min_length:    int
    :param min_length:    (Optional, def=0) Minimum length of random block
    :type  max_length:    int
    :param max_length:    (Optional, def=1) Maximum length of random block
    :type  num_mutations: int
    :param num_mutations: (Optional, def=25) Number of mutations to make before reverting to default
    :type  fuzzable:      bool
    :param fuzzable:      (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  step:          int
    :param step:          (Optional, def=None) If not null, step count between min and max reps, otherwise random
    :type  name:          str
    :param name:          (Optional, def=None) Specifying a name gives you direct access to a primitive
    """

    blocks.CURRENT.push(
        RandomData(
            name=name,
            default_value=value,
            min_length=min_length,
            max_length=max_length,
            max_mutations=num_mutations,
            step=step,
            fuzzable=fuzzable,
        )
    )


def s_static(value=None, name=None):
    """
    Push a static value onto the current block stack.

    :see: Aliases: s_dunno(), s_raw(), s_unknown()

    :type  value: Raw
    :param value: Raw static data
    :type  name:  str
    :param name:  (Optional, def=None) Specifying a name gives you direct access to a primitive
    """

    blocks.CURRENT.push(Static(name=name, default_value=value))


def s_simple(value=None, name=None, fuzz_values=None, fuzzable=True):
    """
    Push a "Simple" primitive onto the current block stack. The only mutations will be those specified  in fuzz_values.

    :see: Aliases: s_dunno(), s_raw(), s_unknown()

    :type  value: Raw
    :param value: Raw static data
    :type  name:  str
    :param name:  (Optional, def=None) Specifying a name gives you direct access to a primitive
    :type  fuzz_values:  list
    :param fuzz_values:  (Optional, def=None) List of fuzz values.
    :type  fuzzable: bool
    :param fuzzable: (Optional, def=True) Enable/disable fuzzing of this primitive
    """

    blocks.CURRENT.push(Simple(name=name, default_value=value, fuzz_values=fuzz_values, fuzzable=fuzzable))


def s_mirror(primitive_name=None, name=None):
    """
    Push a mirror of another primitive onto the current block stack.

    :type primitive_name:   str
    :param primitive_name:  (Optional, def=None) Name of target primitive
    :type name:             str
    :param name:            (Optional, def=None) Name of current primitive
    """
    blocks.CURRENT.push(Mirror(name=name, primitive_name=primitive_name, request=blocks.CURRENT))


def s_string(value="", size=None, padding=b"\x00", encoding="ascii", fuzzable=True, max_len=None, name=None):
    """
    Push a string onto the current block stack.

    :type  value:    str
    :param value:    (Optional, def="")Default string value
    :type  size:     int
    :param size:     (Optional, def=None) Static size of this field, leave None for dynamic.
    :type  padding:  Character
    :param padding:  (Optional, def="\\x00") Value to use as padding to fill static field size.
    :type  encoding: str
    :param encoding: (Optional, def="ascii") String encoding, ex: utf_16_le for Microsoft Unicode.
    :type  fuzzable: bool
    :param fuzzable: (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  max_len:  int
    :param max_len:  (Optional, def=None) Maximum string length
    :type  name:     str
    :param name:     (Optional, def=None) Specifying a name gives you direct access to a primitive
    """
    # support old interface where default was -1 instead of None
    if size == -1:
        size = None
    if max_len == -1:
        max_len = None

    blocks.CURRENT.push(
        String(
            name=name,
            default_value=value,
            size=size,
            padding=padding,
            encoding=encoding,
            max_len=max_len,
            fuzzable=fuzzable,
        )
    )


def s_from_file(value=b"", filename=None, encoding="ascii", fuzzable=True, max_len=0, name=None):
    """
    Push a value from file onto the current block stack.

    :type  value:    bytes
    :param value:    (Optional, def=b"") Default bytes value
    :type  filename: str
    :param filename: (Optional, def=None) Filename pattern to load all fuzz value
    :type  encoding: str
    :param encoding: (DEPRECIATED, def="ascii") String encoding, ex: utf_16_le for Microsoft Unicode.
    :type  fuzzable: bool
    :param fuzzable: (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  max_len:  int
    :param max_len:  (Optional, def=0) Maximum string length
    :type  name:     str
    :param name:     (Optional, def=None) Specifying a name gives you direct access to a primitive
    """

    blocks.CURRENT.push(FromFile(name=name, default_value=value, max_len=max_len, filename=filename, fuzzable=fuzzable))


def s_float(
    value=0.0,
    s_format=".1f",
    f_min=0.0,
    f_max=100.0,
    max_mutations=1000,
    fuzzable=True,
    seed=None,
    encode_as_ieee_754=False,
    endian="big",
    name=None,
):
    """
    Push a float onto the current block stack.

    :type  value:              float
    :param value:              (Optional, def=0.0) Default float value
    :type  s_format:           str
    :param s_format:           (Optional, def=".1f") Format of the float value after encoding.
    :type f_min:               float
    :param f_min:              (Optional, def=0.0) Min value of generated floats.
    :type f_max:               float
    :param f_max:              (Optional, def=100.0) Max value of generated floats.
    :type max_mutations:       int
    :param max_mutations:      (Optional, def=1000) Total number of mutations for this individual primitive
    :type  fuzzable:           bool
    :param fuzzable:           (Optional, def=True) Enable/disable fuzzing of this primitive
    :type seed:                int or str or bytes or bytearray
    :param seed:               (Optional, def=None) Set random.seed() with the given seed for reproducible results
    :type encode_as_ieee_754:  bool
    :param encode_as_ieee_754: (Optional, def=False) Enable/disable encoding as IEEE 754 float
    :type endian:              str
    :param endian:             (Optional, def='big') Set the endianness for IEEE 754 float
    :type  name:               str
    :param name:               (Optional, def=None) Specifying a name gives you direct access to a primitive
    """
    blocks.CURRENT.push(
        Float(
            name=name,
            default_value=value,
            s_format=s_format,
            f_min=f_min,
            f_max=f_max,
            max_mutations=max_mutations,
            fuzzable=fuzzable,
            seed=seed,
            encode_as_ieee_754=encode_as_ieee_754,
            endian=endian,
        )
    )


# noinspection PyTypeChecker
def s_bit_field(
    value=0,
    width=8,
    endian=LITTLE_ENDIAN,
    output_format="binary",
    signed=False,
    full_range=False,
    fuzzable=True,
    name=None,
    fuzz_values=None,
):
    """
    Push a variable length bit field onto the current block stack.

    :see: Aliases: s_bit(), s_bits()

    :type  value:          int
    :param value:          (Optional, def=0) Default integer value
    :type  width:          int
    :param width:          (Optional, def=8) Width of bit fields
    :type  endian:         Character
    :param endian:         (Optional, def=LITTLE_ENDIAN) Endianness of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >)
    :type  output_format:  str
    :param output_format:  (Optional, def=binary) Output format, "binary" or "ascii"
    :type  signed:         bool
    :param signed:         (Optional, def=False) Make size signed vs. unsigned (applicable only with format="ascii")
    :type  full_range:     bool
    :param full_range:     (Optional, def=False) If enabled the field mutates through *all* possible values.
    :type  fuzzable:       bool
    :param fuzzable:       (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  name:           str
    :param name:           (Optional, def=None) Specifying a name gives you direct access to a primitive
    :type fuzz_values:     list
    :param fuzz_values:    List of custom fuzz values to add to the normal mutations.
    """

    blocks.CURRENT.push(
        BitField(
            name=name,
            default_value=value,
            width=width,
            endian=endian,
            output_format=output_format,
            signed=signed,
            full_range=full_range,
            fuzzable=fuzzable,
            fuzz_values=fuzz_values,
        )
    )


def s_byte(
    value=0,
    endian=LITTLE_ENDIAN,
    output_format="binary",
    signed=False,
    full_range=False,
    fuzzable=True,
    name=None,
    fuzz_values=None,
):
    """
    Push a byte onto the current block stack.

    :see: Aliases: s_char()

    :type  value:         int|byte
    :param value:         (Optional, def=0) Default integer value
    :type  endian:        Character
    :param endian:        (Optional, def=LITTLE_ENDIAN) Endianess of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >)
    :type  output_format: str
    :param output_format: (Optional, def=binary) Output format, "binary" or "ascii"
    :type  signed:        bool
    :param signed:        (Optional, def=False) Make size signed vs. unsigned (applicable only with format="ascii")
    :type  full_range:    bool
    :param full_range:    (Optional, def=False) If enabled the field mutates through *all* possible values.
    :type  fuzzable:      bool
    :param fuzzable:      (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  name:          str
    :param name:          (Optional, def=None) Specifying a name gives you direct access to a primitive
    :type fuzz_values:    list
    :param fuzz_values:   List of custom fuzz values to add to the normal mutations.
    """

    blocks.CURRENT.push(
        Byte(
            endian=endian,
            output_format=output_format,
            signed=signed,
            full_range=full_range,
            name=name,
            default_value=value,
            fuzzable=fuzzable,
            fuzz_values=fuzz_values,
        )
    )


def s_bytes(value=b"", size=None, padding=b"\x00", fuzzable=True, max_len=None, name=None):
    """
    Push a bytes field of arbitrary length onto the current block stack.

    :type  value:        bytes
    :param value:        (Optional, def=b"")Default binary value
    :type  size:         int
    :param size:         (Optional, def=None) Static size of this field, leave None for dynamic.
    :type  padding:      chr
    :param padding:      (Optional, def=b"\\x00") Value to use as padding to fill static field size.
    :type  fuzzable:     bool
    :param fuzzable:     (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  max_len:      int
    :param max_len:      (Optional, def=None) Maximum string length
    :type  name:         str
    :param name:         (Optional, def=None) Specifying a name gives you direct access to a primitive
    """

    blocks.CURRENT.push(
        Bytes(name=name, default_value=value, size=size, padding=padding, max_len=max_len, fuzzable=fuzzable)
    )


def s_word(
    value=0,
    endian=LITTLE_ENDIAN,
    output_format="binary",
    signed=False,
    full_range=False,
    fuzzable=True,
    name=None,
    fuzz_values=None,
):
    """
    Push a word onto the current block stack.

    :see: Aliases: s_short()

    :type  value:         (Optional, def=0) int
    :param value:         Default integer value
    :type  endian:        chr
    :param endian:        (Optional, def=LITTLE_ENDIAN) Endianess of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >)
    :type  output_format: str
    :param output_format: (Optional, def=binary) Output format, "binary" or "ascii"
    :type  signed:        bool
    :param signed:        (Optional, def=False) Make size signed vs. unsigned (applicable only with format="ascii")
    :type  full_range:    bool
    :param full_range:    (Optional, def=False) If enabled the field mutates through *all* possible values.
    :type  fuzzable:      bool
    :param fuzzable:      (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  name:          str
    :param name:          (Optional, def=None) Specifying a name gives you direct access to a primitive
    :type fuzz_values:    list
    :param fuzz_values:   List of custom fuzz values to add to the normal mutations.
    """

    blocks.CURRENT.push(
        Word(
            endian=endian,
            output_format=output_format,
            signed=signed,
            full_range=full_range,
            name=name,
            default_value=value,
            fuzzable=fuzzable,
            fuzz_values=fuzz_values,
        )
    )


def s_dword(
    value=0,
    endian=LITTLE_ENDIAN,
    output_format="binary",
    signed=False,
    full_range=False,
    fuzzable=True,
    name=None,
    fuzz_values=None,
):
    """
    Push a double word onto the current block stack.

    :see: Aliases: s_long(), s_int()

    :type  value:         (Optional, def=0) int
    :param value:         Default integer value
    :type  endian:        Character
    :param endian:        (Optional, def=LITTLE_ENDIAN) Endianess of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >)
    :type  output_format: str
    :param output_format: (Optional, def=binary) Output format, "binary" or "ascii"
    :type  signed:        bool
    :param signed:        (Optional, def=False) Make size signed vs. unsigned (applicable only with format="ascii")
    :type  full_range:    bool
    :param full_range:    (Optional, def=False) If enabled the field mutates through *all* possible values.
    :type  fuzzable:      bool
    :param fuzzable:      (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  name:          str
    :param name:          (Optional, def=None) Specifying a name gives you direct access to a primitive
    :type fuzz_values:    list
    :param fuzz_values:   List of custom fuzz values to add to the normal mutations.
    """

    blocks.CURRENT.push(
        DWord(
            endian=endian,
            output_format=output_format,
            signed=signed,
            full_range=full_range,
            name=name,
            default_value=value,
            fuzzable=fuzzable,
            fuzz_values=fuzz_values,
        )
    )


def s_qword(
    value=0,
    endian=LITTLE_ENDIAN,
    output_format="binary",
    signed=False,
    full_range=False,
    fuzzable=True,
    name=None,
    fuzz_values=None,
):
    """
    Push a quad word onto the current block stack.

    :see: Aliases: s_double()

    :type  value:         (Optional, def=0) int
    :param value:         Default integer value
    :type  endian:        Character
    :param endian:        (Optional, def=LITTLE_ENDIAN) Endianess of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >)
    :type  output_format: str
    :param output_format: (Optional, def=binary) Output format, "binary" or "ascii"
    :type  signed:        bool
    :param signed:        (Optional, def=False) Make size signed vs. unsigned (applicable only with format="ascii")
    :type  full_range:    bool
    :param full_range:    (Optional, def=False) If enabled the field mutates through *all* possible values.
    :type  fuzzable:      bool
    :param fuzzable:      (Optional, def=True) Enable/disable fuzzing of this primitive
    :type  name:          str
    :param name:          (Optional, def=None) Specifying a name gives you direct access to a primitive
    :type fuzz_values:    list
    :param fuzz_values:   List of custom fuzz values to add to the normal mutations.
    """

    blocks.CURRENT.push(
        QWord(
            endian=endian,
            output_format=output_format,
            signed=signed,
            full_range=full_range,
            name=name,
            default_value=value,
            fuzzable=fuzzable,
            fuzz_values=fuzz_values,
        )
    )


# ALIASES


s_dunno = s_raw = s_unknown = s_static
s_sizer = s_size
s_bit = s_bits = s_bit_field
s_char = s_byte
s_short = s_word
s_long = s_int = s_dword
s_double = s_qword
s_repeater = s_repeat


def s_intelword(*args, **kwargs):
    defaults = {"endian": "LITTLE_ENDIAN"}
    defaults.update(kwargs)
    return s_long(*args, **defaults)


def s_intelhalfword(*args, **kwargs):
    defaults = {"endian": "LITTLE_ENDIAN"}
    defaults.update(kwargs)
    return s_short(*args, **defaults)


def s_bigword(*args, **kwargs):
    defaults = {"endian": "BIG_ENDIAN"}
    defaults.update(kwargs)
    return s_long(*args, **defaults)


def s_cstring(x):
    s_string(x)
    s_static("\x00")


# Not implemented aliases yet
def not_impl(alias, *args, **kwargs):
    raise NotImplementedError("%s isn't implemented yet. Args -> %s. Kwargs -> %s" % (alias, args, kwargs))


s_string_lf = functools.partial(not_impl, "s_string_lf")
s_string_or_env = functools.partial(not_impl, "s_string_or_env")
s_string_repeat = functools.partial(not_impl, "s_string_repeat")
s_string_variable = functools.partial(not_impl, "s_string_variable")
s_string_variables = functools.partial(not_impl, "s_string_variables")
s_binary_repeat = functools.partial(not_impl, "s_binary_repeat")
s_unistring_variable = functools.partial(not_impl, "s_unistring_variable")
s_xdr_string = functools.partial(not_impl, "s_xdr_string")


def no_sizer(*args, **kwargs):
    _ = kwargs  # just making the function univesral
    raise exception.SizerNotUtilizedError("Use the s_size primitive for including sizes. Args -> %s" % args)


# A bunch of un-defined primitives from SPIKE
s_binary_block_size_intel_halfword_plus_variable = no_sizer
s_binary_block_size_halfword_bigendian_variable = no_sizer
s_binary_block_size_word_bigendian_plussome = no_sizer
s_binary_block_size_word_bigendian_variable = no_sizer
s_binary_block_size_halfword_bigendian_mult = no_sizer
s_binary_block_size_intel_halfword_variable = no_sizer
s_binary_block_size_intel_halfword_mult = no_sizer
s_binary_block_size_intel_halfword_plus = no_sizer
s_binary_block_size_halfword_bigendian = no_sizer
s_binary_block_size_word_intel_mult_plus = no_sizer
s_binary_block_size_intel_word_variable = no_sizer
s_binary_block_size_word_bigendian_mult = no_sizer
s_blocksize_unsigned_string_variable = no_sizer
s_binary_block_size_intel_word_plus = no_sizer
s_binary_block_size_intel_halfword = no_sizer
s_binary_block_size_word_bigendian = no_sizer
s_blocksize_signed_string_variable = no_sizer
s_binary_block_size_byte_variable = no_sizer
s_binary_block_size_intel_word = no_sizer
s_binary_block_size_byte_plus = no_sizer
s_binary_block_size_byte_mult = no_sizer
s_blocksize_asciihex_variable = no_sizer
s_binary_block_size_byte = no_sizer
s_blocksize_asciihex = no_sizer
s_blocksize_string = no_sizer


# MISC


def s_hex_dump(data, addr=0):
    """
    Return the hex dump of the supplied data starting at the offset address specified.

    :type  data: Raw
    :param data: Data to show hex dump of
    :type  addr: int
    :param addr: (Optional, def=0) Offset to start displaying hex dump addresses from

    :rtype:  str
    :return: Hex dump of raw data
    """

    dump = byte_slice = ""

    for byte in data:
        if addr % 16 == 0:
            dump += " "

            for char in byte_slice:
                if 32 <= ord(char) <= 126:
                    dump += char
                else:
                    dump += "."

            dump += "\n%04x: " % addr
            byte_slice = ""

        dump += "%02x " % ord(byte)
        byte_slice += byte
        addr += 1

    remainder = addr % 16

    if remainder != 0:
        dump += "   " * (16 - remainder) + " "

    for char in byte_slice:
        if 32 <= ord(char) <= 126:
            dump += char
        else:
            dump += "."

    return dump + "\n"



#!/usr/bin/env python3
# Designed for use with boofuzz v0.2.0

# Original author:
#
# pedram amini <pamini@tippingpoint.com>
#
# on vmware:
#     cd shared\boofuzz\branches\pedram
#     network_monitor.py -d 1 -f "src or dst port 5298" -p audits\trillian_jabber
#     process_monitor.py -c audits\trillian_jabber.crashbin -p trillian.exe
#
# on localhost:
#     vmcontrol.py -r "c:\Progra~1\VMware\VMware~1\vmrun.exe" \
#                  -x "v:\vmfarm\images\windows\xp\win_xp_pro-clones\allsor~1\win_xp_pro.vmx" \
#                  --snapshot "boofuzz ready and waiting"
#
# note:
#     you MUST register the IP address of the fuzzer as a valid MDNS "presence" host. to do so, simply install and
#     launch trillian on the fuzz box with rendezvous enabled. otherwise the target will drop the connection.
#

from boofuzz import pedrpc, s_get, sessions, TCPSocketConnection

# noinspection PyUnresolvedReferences
# pytype: disable=import-error
from request_definitions import jabber  # noqa: F401

# pytype: enable=import-error


def init_message(sock):
    init = '<?xml version="1.0" encoding="UTF-8" ?>\n'
    init += '<stream:stream to="152.67.137.126" xmlns="jabber:client" xmlns:stream="http://etherx.jabber.org/streams">'

    sock.send(init)
    sock.recv(1024)


sess = sessions.Session(session_filename="audits/trillian.session")
target = sessions.Target(connection=TCPSocketConnection("152.67.137.126", 5298))
target.netmon = pedrpc.Client("152.67.137.126", 26001)
target.procmon = pedrpc.Client("152.67.137.126", 26002)
target.vmcontrol = pedrpc.Client("127.0.0.1", 26003)
target.procmon_options = {"proc_name": "trillian.exe"}

# start up the target.
target.vmcontrol.restart_target()
print("virtual machine up and running")

sess.add_target(target)
sess.pre_send = init_message
sess.connect(sess.root, s_get("chat message"))
sess.fuzz()



import collections

from .. import exception
from ..constants import ERR_NAME_NO_RESOLVE, ERR_NAME_NOT_FOUND, ERR_NAME_TOO_MANY
from ..exception import BoofuzzNameResolutionError
from ..fuzzable import Fuzzable
from ..fuzzable_block import FuzzableBlock
from ..pgraph.node import Node


class Request(FuzzableBlock, Node):
    """Top level container. Can hold any block structure or primitive.

    This can essentially be thought of as a super-block, root-block, daddy-block or whatever other alias you prefer.

    :param name: Name of this request
    :type name: str, optional
    :param children: Children of this request, defaults to None
    :type children: boofuzz.Fuzzable, optional
    """

    def __init__(self, name=None, children=None):
        FuzzableBlock.__init__(self, name=name, request=self)
        Node.__init__(self)
        self.label = name  # node label for graph rendering.
        self.stack = []  # the request stack.
        self.block_stack = []  # list of open blocks, -1 is last open block.
        self.callbacks = collections.defaultdict(list)
        self.names = {name: self}  # dictionary of directly accessible primitives.
        self._rendered = b""  # rendered block structure.
        self._mutant_index = 0  # current mutation index.
        self._element_mutant_index = None  # index of current mutant element within self.stack
        self.mutant = None  # current primitive being mutated.

        if children is None:
            children = []
        elif isinstance(children, Fuzzable):
            children = [children]

        self._initialize_children(child_nodes=children)

    def _initialize_children(self, child_nodes, block_stack=None):
        if block_stack is None:
            block_stack = list()

        for item in child_nodes:
            item.context_path = self._generate_context_path(block_stack)
            item.request = self
            # ensure the name doesn't already exist.
            if item.qualified_name in list(self.names):
                raise exception.SullyRuntimeError("BLOCK NAME ALREADY EXISTS: %s" % item.qualified_name)
            self.names[item.qualified_name] = item

            if len(block_stack) == 0:
                self.stack.append(item)
            if isinstance(item, FuzzableBlock):
                block_stack.append(item)
                self._initialize_children(child_nodes=item.stack, block_stack=block_stack)
                block_stack.pop()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def fuzzable(self):
        return True

    def pop(self):
        """
        The last open block was closed, so pop it off of the block stack.
        """

        if not self.block_stack:
            raise exception.SullyRuntimeError("BLOCK STACK OUT OF SYNC")

        self.block_stack.pop()

    def push(self, item):
        """
        Push an item into the block structure. If no block is open, the item goes onto the request stack. otherwise,
        the item goes onto the last open blocks stack.

        What this method does:
        1. Sets context_path for each pushed FuzzableWrapper.
        2. Sets request for each FuzzableWrapper
        3. Checks for duplicate qualified_name items
        4. Adds item to self.names map (based on qualified_name)
        5. Adds the item to self.stack, or to the stack of the currently opened block.

        Also: Manages block_stack, mostly an implementation detail to help static protocol definition

        @type item: BasePrimitive | Block | Request | Size | Repeat
        @param item: Some primitive/block/request/etc.
        """
        item.context_path = self._generate_context_path(self.block_stack)
        item.request = self
        # ensure the name doesn't already exist.
        if item.qualified_name in list(self.names):
            raise exception.SullyRuntimeError("BLOCK NAME ALREADY EXISTS: %s" % item.qualified_name)

        self.names[item.qualified_name] = item

        # if there are no open blocks, the item gets pushed onto the request stack.
        # otherwise, the pushed item goes onto the stack of the last opened block.
        if not self.block_stack:
            self.stack.append(item)
        else:
            self.block_stack[-1].push(item)

        # add the opened block to the block stack.
        if isinstance(item, FuzzableBlock):
            self.block_stack.append(item)

    def _generate_context_path(self, block_stack):
        context_path = ".".join(x.name for x in block_stack)  # TODO put in method
        context_path = ".".join(filter(None, (self.name, context_path)))
        return context_path

    def render(self, mutation_context=None):
        if self.block_stack:
            raise exception.SullyRuntimeError("UNCLOSED BLOCK: %s" % self.block_stack[-1].qualified_name)

        return self.get_child_data(mutation_context=mutation_context)

    def walk(self, stack=None):
        """
        Recursively walk through and yield every primitive and block on the request stack.

        @param stack: Set to none -- used internally by recursive calls.
                      If None, uses self.stack.

        @rtype:  Sulley Primitives
        @return: Sulley Primitives
        """

        if not stack:
            stack = self.stack

        for item in stack:
            # if the item is a block, step into it and continue looping.
            if isinstance(item, FuzzableBlock):
                for stack_item in self.walk(item.stack):
                    yield stack_item
            else:
                yield item

    def resolve_name(self, context_path, name):
        """
        Names are resolved thus:
        1. If the name starts with a dot, it is treated as a relative path name in the style of PEP 328.
            1. "." refers to the current directory, so to speak.
            2. ".." refers to the next directory up.
            3. "..." refers to another directory up, and so forth.
        2. If the name does _not_ start with a dot, it is treated as an absolute name.
        3. Backwards compatibility: If the absolute name fails to resolve, the engine searches for any block or
            primitive with that name. If more or less than exactly one match is found, an error results.

        Args:
            context_path: The "current working directory" for resolving the name. E.g. "block_1.block_2".
            name: The name being resolved. May be absolute or relative.

        Returns:

        """
        if name is None:
            raise BoofuzzNameResolutionError(ERR_NAME_NOT_FOUND.format(name))
        if name.startswith("."):  # Case 1 relative
            components = (context_path + name).split(".")  # double dots leave an empty string; so do trailing dots
            while "" in components:
                i = components.index("")
                if i <= 0:
                    raise BoofuzzNameResolutionError(ERR_NAME_NO_RESOLVE.format(name, context_path))
                elif i == len(components) - 1:  # last in list; indicates a trailing dot
                    del components[i]
                else:  # double dot
                    del components[i]
                    del components[i - 1]
            return self._lookup_resolved_name(".".join(components))
        else:
            full_absolute_name = "{0}.{1}".format(self._name, name)
            if full_absolute_name in self.names:  # Case 2 absolute
                return self._lookup_resolved_name(full_absolute_name)
            else:  # Case 3 backwards compatibility --  look up by last name component
                found_names = [n for n in self.names if n.rsplit(".")[-1] == name]
                if len(found_names) == 1:
                    return self.names[found_names[0]]
                elif len(found_names) == 0:
                    raise BoofuzzNameResolutionError(ERR_NAME_NOT_FOUND.format(name))
                else:
                    raise BoofuzzNameResolutionError(ERR_NAME_TOO_MANY.format(name, found_names))

    def _lookup_resolved_name(self, resolved_name):
        if resolved_name in self.names:
            return self.names[resolved_name]
        else:
            raise BoofuzzNameResolutionError(ERR_NAME_NOT_FOUND.format(resolved_name))

    def get_mutations(self, default_value=None, skip_elements=None):
        return self.mutations(default_value=default_value, skip_elements=skip_elements)

    def get_num_mutations(self):
        return self.num_mutations()

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.name)



#
# pGRAPH
# Copyright (C) 2006 Pedram Amini <pedram.amini@gmail.com>
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

import pydot


class Node:
    id = 0
    number = 0

    # general graph attributes
    color = 0xEEF7FF
    border_color = 0xEEEEEE
    label = ""
    shape = "box"

    # gml relevant attributes.
    gml_width = 0.0
    gml_height = 0.0
    gml_pattern = "1"
    gml_stipple = 1
    gml_line_width = 1.0
    gml_type = "rectangle"
    gml_width_shape = 1.0

    # udraw relevant attributes.
    udraw_image = None
    udraw_info = ""

    def __init__(self, node_id=None):
        self.id = node_id
        self.number = 0

        # general graph attributes
        self.color = 0xEEF7FF
        self.border_color = 0xEEEEEE
        self.label = ""
        self.shape = "box"

        # gml relevant attributes.
        self.gml_width = 0.0
        self.gml_height = 0.0
        self.gml_pattern = "1"
        self.gml_stipple = 1
        self.gml_line_width = 1.0
        self.gml_type = "rectangle"
        self.gml_width_shape = 1.0

    def render_node_gml(self):
        """
        Render a node description suitable for use in a GML file using the set internal attributes.

        @rtype:  String
        @return: GML node description.
        """

        # GDE does not like lines longer then approx 250 bytes. within their their own GML files you won't find lines
        # longer then approx 210 bytes. wo we are forced to break long lines into chunks.
        chunked_label = ""
        cursor = 0

        while cursor < len(self.label):
            amount = 200

            # if the end of the current chunk contains a backslash or double-quote, back off some.
            if cursor + amount < len(self.label):
                while self.label[cursor + amount] == "\\" or self.label[cursor + amount] == '"':
                    amount -= 1

            chunked_label += self.label[cursor : cursor + amount] + "\\\n"
            cursor += amount

        # if node width and height were not explicitly specified, make a best effort guess to create something nice.
        if not self.gml_width:
            self.gml_width = len(self.label) * 10

        if not self.gml_height:
            self.gml_height = len(self.label.split()) * 20

        # construct the node definition.
        node = (
            "  node [\n"
            "    id {number}\n"
            '    template "oreas:std:rect"\n'
            '    label "<!--{id:08x}--> {chunked_label}"\n'
            "    graphics [\n"
            "      w {gml_width}\n"
            "      h {gml_height}\n"
            '      fill "#{color:06x}"\n'
            '      line "#{border_color:06x}"\n'
            '      pattern "{gml_pattern}"\n'
            "      stipple {gml_stipple}\n"
            "      lineWidth {gml_line_width}\n"
            '      type "{gml_type}"\n'
            "      width {gml_width_shape}\n"
            "    ]\n"
            "  ]\n".format(
                number=self.number,
                id=self.id,
                gml_width=self.gml_width,
                gml_height=self.gml_height,
                color=self.color,
                border_color=self.border_color,
                gml_pattern=self.gml_pattern,
                gml_stipple=self.gml_stipple,
                gml_line_width=self.gml_line_width,
                gml_type=self.gml_type,
                gml_width_shape=self.gml_width_shape,
                chunked_label=chunked_label,
            )
        )

        return node

    def render_node_graphviz(self):
        """
        Render a node suitable for use in a Pydot graph using the set internal attributes.

        @rtype:  pydot.Node
        @return: Pydot object representing node
        """

        dot_node = pydot.Node(self.id)

        dot_node.obj_dict["attributes"]["label"] = '<<font face="lucida console">{}</font>>'.format(
            self.label.rstrip("\r\n")
        )
        dot_node.obj_dict["attributes"]["label"] = dot_node.obj_dict["attributes"]["label"].replace("\\n", "<br/>")
        dot_node.obj_dict["attributes"]["shape"] = self.shape
        dot_node.obj_dict["attributes"]["color"] = "#{:06x}".format(self.color)
        dot_node.obj_dict["attributes"]["fillcolor"] = "#{:06x}".format(self.color)

        return dot_node

    def render_node_udraw(self, graph):
        """
        Render a node description suitable for use in a uDraw file using the set internal attributes.

        @type  graph: pgraph.Graph
        @param graph: Top level graph object containing the current node

        @rtype:  String
        @return: uDraw node description.
        """

        # translate newlines for uDraw.
        self.label = self.label.replace("\n", "\\n")

        # if an image was specified for this node, update the shape and include the image tag.
        if self.udraw_image:
            self.shape = "image"
            udraw_image = 'a("IMAGE","{}"),'.format(self.udraw_image)
        else:
            udraw_image = ""

        udraw = (
            '\n  l("{id:08x}",\n'
            '    n("",\n'
            "      [\n"
            "        {udraw_image}\n"
            '        a("_GO","{shape}"),\n'
            '        a("COLOR","#{color:08x}"),\n'
            '        a("OBJECT","{label}"),\n'
            '        a("FONTFAMILY","courier"),\n'
            '        a("INFO","{udraw_info}"),\n'
            '        a("BORDER","none")\n'
            "      ]\n"
            "      [\n".format(
                id=self.id,
                udraw_image=udraw_image,
                shape=self.shape,
                color=self.color,
                label=self.label,
                udraw_info=self.udraw_info,
            )
        )

        edges = graph.edges_from(self.id)

        for edge in edges:
            udraw += edge.render_edge_udraw(graph)
            udraw += ","

        if edges:
            udraw = udraw[0:-1]

        udraw += "  ]))"

        return udraw

    def render_node_udraw_update(self):
        """
        Render a node update description suitable for use in a uDraw file using the set internal attributes.

        @rtype:  String
        @return: uDraw node update description.
        """

        # translate newlines for uDraw.
        self.label = self.label.replace("\n", "\\n")

        # if an image was specified for this node, update the shape and include the image tag.
        if self.udraw_image:
            self.shape = "image"
            udraw_image = 'a("IMAGE","{}"),'.format(self.udraw_image)
        else:
            udraw_image = ""

        udraw = (
            '\n  new_node("{id:08x}","",\n'
            "    ["
            "      {udraw_image}\n"
            '      a("_GO","{shape}"),'
            '      a("COLOR","#{color:08x}"),\n'
            '      a("OBJECT","{label}"),\n'
            '      a("FONTFAMILY","courier"),\n'
            '      a("INFO","{udraw_info}"),\n'
            '      a("BORDER","none")\n'
            "    ]\n"
            "  )\n".format(
                id=self.id,
                udraw_image=udraw_image,
                shape=self.shape,
                color=self.color,
                label=self.label,
                udraw_info=self.udraw_info,
            )
        )

        return udraw



#
# Crash Binning
# Copyright (C) 2006 Pedram Amini <pedram.amini@gmail.com>
#
# $Id: crash_binning.py 193 2007-04-05 13:30:01Z cameron $
#
# This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with this program; if not, write to the Free
# Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
#

"""
@author:       Pedram Amini
@license:      GNU General Public License 2.0 or later
@contact:      pedram.amini@gmail.com
@organization: www.openrce.org
"""

import json
from io import open


class CrashBinStruct:
    def __init__(self):
        self.exception_module = None
        self.exception_address = 0
        self.write_violation = 0
        self.violation_address = 0
        self.violation_thread_id = 0
        self.context = None
        self.context_dump = None
        self.disasm = None
        self.disasm_around = []
        self.stack_unwind = []
        self.seh_unwind = []
        self.extra = None


class CrashBinning:
    """
    @todo: Add MySQL import/export.
    """

    bins = {}
    last_crash = None
    pydbg = None

    def __init__(self):
        self.bins = {}
        self.last_crash = None
        self.pydbg = None

    def record_crash(self, pydbg, extra=None):
        """
        Given a PyDbg instantiation that at the current time is assumed to have "crashed" (access violation for example)
        record various details such as the disassemly around the violating address, the ID of the offending thread, the
        call stack and the SEH unwind. Store the recorded data in an internal dictionary, binning them by the exception
        address.

        @type  pydbg: pydbg
        @param pydbg: Instance of pydbg
        @type  extra: Mixed
        @param extra: (Optional, Def=None) Whatever extra data you want to store with this bin
        """

        self.pydbg = pydbg
        crash = CrashBinStruct()

        # add module name to the exception address.
        exception_module = pydbg.addr_to_module(pydbg.dbg.u.Exception.ExceptionRecord.ExceptionAddress)

        if exception_module:
            exception_module = exception_module.szModule
        else:
            exception_module = "[INVALID]"

        crash.exception_module = exception_module
        crash.exception_address = pydbg.dbg.u.Exception.ExceptionRecord.ExceptionAddress
        crash.write_violation = pydbg.dbg.u.Exception.ExceptionRecord.ExceptionInformation[0]
        crash.violation_address = pydbg.dbg.u.Exception.ExceptionRecord.ExceptionInformation[1]
        crash.violation_thread_id = pydbg.dbg.dwThreadId
        crash.context = pydbg.context
        crash.context_dump = pydbg.dump_context(pydbg.context, print_dots=False)
        crash.disasm = pydbg.disasm(crash.exception_address)
        crash.disasm_around = pydbg.disasm_around(crash.exception_address, 10)
        crash.stack_unwind = pydbg.stack_unwind()
        crash.seh_unwind = pydbg.seh_unwind()
        crash.extra = extra

        # add module names to the stack unwind.
        for i in range(len(crash.stack_unwind)):
            addr = crash.stack_unwind[i]
            module = pydbg.addr_to_module(addr)

            if module:
                module = module.szModule
            else:
                module = "[INVALID]"

            crash.stack_unwind[i] = "%s:%08x" % (module, addr)

        # add module names to the SEH unwind.
        for i in range(len(crash.seh_unwind)):
            (addr, handler) = crash.seh_unwind[i]

            module = pydbg.addr_to_module(handler)

            if module:
                module = module.szModule
            else:
                module = "[INVALID]"

            crash.seh_unwind[i] = (addr, handler, "%s:%08x" % (module, handler))

        if crash.exception_address not in self.bins:
            self.bins[crash.exception_address] = []

        self.bins[crash.exception_address].append(crash)
        self.last_crash = crash

    def crash_synopsis(self, crash=None):
        """
        For the supplied crash, generate and return a report containing the disassemly around the violating address,
        the ID of the offending thread, the call stack and the SEH unwind. If not crash is specified, then call through
        to last_crash_synopsis() which returns the same information for the last recorded crash.

        @see: crash_synopsis()

        @type  crash: CrashBinStruct
        @param crash: (Optional, def=None) Crash object to generate report on

        @rtype:  str
        @return: Crash report
        """

        if not crash:
            return self.last_crash_synopsis()

        if crash.write_violation:
            direction = "write to"
        else:
            direction = "read from"

        synopsis = "%s:%08x %s from thread %d caused access violation\nwhen attempting to %s 0x%08x\n\n" % (
            crash.exception_module,
            crash.exception_address,
            crash.disasm,
            crash.violation_thread_id,
            direction,
            crash.violation_address,
        )

        synopsis += crash.context_dump

        synopsis += "\ndisasm around:\n"
        for ea, inst in crash.disasm_around:
            synopsis += "\t0x%08x %s\n" % (ea, inst)

        if len(crash.stack_unwind):
            synopsis += "\nstack unwind:\n"
            for entry in crash.stack_unwind:
                synopsis += "\t%s\n" % entry

        if len(crash.seh_unwind):
            synopsis += "\nSEH unwind:\n"
            for addr, handler, handler_str in crash.seh_unwind:
                synopsis += "\t%08x -> %s\n" % (addr, handler_str)

        return synopsis + "\n"

    def export_file(self, file_name):
        """
        Dump the entire object structure to disk.

        @see: import_file()

        @type  file_name:   str
        @param file_name:   File name to export to

        @rtype:             CrashBinning
        @return:            self
        """

        # null out what we don't serialize but save copies to restore after dumping to disk.
        last_crash = self.last_crash
        pydbg = self.pydbg

        self.last_crash = self.pydbg = None

        json.dump(self.bins, open(file_name, "wb+"), default=lambda o: o.__dict__)

        self.last_crash = last_crash
        self.pydbg = pydbg

        return self

    def import_file(self, file_name):
        """
        Load the entire object structure from disk.

        @see: export_file()

        @type  file_name:   str
        @param file_name:   File name to import from

        @rtype:             CrashBinning
        @return:            self
        """

        self.bins = {}
        bin_dict = json.load(open(file_name, "rb"))
        for crash_address, bin_list in bin_dict.items():
            self.bins[crash_address] = []
            for single_bin in bin_list:
                tmp = CrashBinStruct()
                tmp.__dict__ = single_bin
                self.bins[crash_address].append(tmp)

        return self

    def last_crash_synopsis(self):
        """
        For the last recorded crash, generate and return a report containing the disassemly around the violating
        address, the ID of the offending thread, the call stack and the SEH unwind.

        @see: crash_synopsis()

        @rtype:  String
        @return: Crash report
        """

        if self.last_crash.write_violation:
            direction = "write to"
        else:
            direction = "read from"

        synopsis = "%s:%08x %s from thread %d caused access violation\nwhen attempting to %s 0x%08x\n\n" % (
            self.last_crash.exception_module,
            self.last_crash.exception_address,
            self.last_crash.disasm,
            self.last_crash.violation_thread_id,
            direction,
            self.last_crash.violation_address,
        )

        synopsis += self.last_crash.context_dump

        synopsis += "\ndisasm around:\n"
        for ea, inst in self.last_crash.disasm_around:
            synopsis += "\t0x%08x %s\n" % (ea, inst)

        if len(self.last_crash.stack_unwind):
            synopsis += "\nstack unwind:\n"
            for entry in self.last_crash.stack_unwind:
                synopsis += "\t%s\n" % entry

        if len(self.last_crash.seh_unwind):
            synopsis += "\nSEH unwind:\n"
            for addr, handler, handler_str in self.last_crash.seh_unwind:
                try:
                    disasm = self.pydbg.disasm(handler)
                except Exception:
                    disasm = "[INVALID]"

                synopsis += "\t%08x -> %s %s\n" % (addr, handler_str, disasm)

        return synopsis + "\n"



import attr

from . import helpers


@attr.s
class DataTestCase:
    name = attr.ib()
    index = attr.ib()
    timestamp = attr.ib()
    steps = attr.ib(default=[])

    @property
    def text_render(self):
        s = helpers.format_log_msg(
            msg_type="test_case", description=self.description, timestamp=self.timestamp, format_type="terminal"
        )
        return s

    @property
    def html_log_line(self):
        return helpers.format_log_msg(
            msg_type="test_case", description=self.description, timestamp=self.timestamp, format_type="html"
        )

    @property
    def description(self):
        return "{index}: {name}".format(index=self.index, name=self.name)

    @property
    def css_class(self):
        return helpers.test_step_info["test_case"]["css_class"]



from .base_primitive import BasePrimitive
from .. import helpers


class Delim(BasePrimitive):
    r"""Represent a delimiter such as :,\r,\n, ,=,>,< etc... Mutations include repetition, substitution and exclusion.

    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type name: str, optional
    :param default_value: Value used when the element is not being fuzzed - should typically represent a valid value.
    :type default_value: char, optional
    :param fuzzable: Enable/disable fuzzing of this primitive, defaults to true
    :type fuzzable: bool, optional
    """

    def __init__(self, name=None, default_value=" ", *args, **kwargs):
        super(Delim, self).__init__(name=name, default_value=default_value, *args, **kwargs)

        self._fuzz_library.append(self._default_value * 2)
        self._fuzz_library.append(self._default_value * 5)
        self._fuzz_library.append(self._default_value * 10)
        self._fuzz_library.append(self._default_value * 25)
        self._fuzz_library.append(self._default_value * 100)
        self._fuzz_library.append(self._default_value * 500)
        self._fuzz_library.append(self._default_value * 1000)

        self._fuzz_library.append("")
        if self._default_value == " ":
            self._fuzz_library.append("\t")
            self._fuzz_library.append("\t" * 2)
            self._fuzz_library.append("\t" * 100)

        self._fuzz_library.append(" ")
        self._fuzz_library.append("\t")
        self._fuzz_library.append("\t " * 100)
        self._fuzz_library.append("\t\r\n" * 100)
        self._fuzz_library.append("!")
        self._fuzz_library.append("@")
        self._fuzz_library.append("#")
        self._fuzz_library.append("$")
        self._fuzz_library.append("%")
        self._fuzz_library.append("^")
        self._fuzz_library.append("&")
        self._fuzz_library.append("*")
        self._fuzz_library.append("(")
        self._fuzz_library.append(")")
        self._fuzz_library.append("{")
        self._fuzz_library.append("}")
        self._fuzz_library.append("[")
        self._fuzz_library.append("]")
        self._fuzz_library.append("-")
        self._fuzz_library.append("_")
        self._fuzz_library.append("+")
        self._fuzz_library.append("=")
        self._fuzz_library.append(":")
        self._fuzz_library.append(": " * 100)
        self._fuzz_library.append(":7" * 100)
        self._fuzz_library.append(";")
        self._fuzz_library.append("'")
        self._fuzz_library.append('"')
        self._fuzz_library.append("/")
        self._fuzz_library.append("\\")
        self._fuzz_library.append("?")
        self._fuzz_library.append("<")
        self._fuzz_library.append(">")
        self._fuzz_library.append(".")
        self._fuzz_library.append(",")
        self._fuzz_library.append("\r")
        self._fuzz_library.append("\n")
        self._fuzz_library.append("\r\n" * 64)
        self._fuzz_library.append("\r\n" * 128)
        self._fuzz_library.append("\r\n" * 512)

    def encode(self, value, mutation_context):
        if value is None:
            value = b""
        return helpers.str_to_bytes(value)



import os

try:
    import resource  # Linux only

    resource.setrlimit(  # Equivalent to: ulimit -c unlimited
        resource.RLIMIT_CORE, (resource.RLIM_INFINITY, resource.RLIM_INFINITY)
    )
except ImportError:
    pass
import signal
import subprocess
import sys
import threading
import time

import psutil
from io import open

if not getattr(__builtins__, "WindowsError", None):

    class WindowsError(OSError):
        """Mock WindowsError since Linux Python lacks WindowsError"""

        @property
        def winerror(self):
            return self.errno

        pass


POPEN_COMMUNICATE_TIMEOUT_FOR_ALREADY_DEAD_TASK = 30


def _enumerate_processes():
    for pid in psutil.pids():
        yield (pid, psutil.Process(pid).name())


def _get_coredump_path():
    """
    This method returns the path to the coredump file if one was created
    """
    if sys.platform == "linux" or sys.platform == "linux2":
        path = "./core"
        if os.path.isfile(path):
            return path

    return None


class DebuggerThreadSimple(threading.Thread):
    """Simple debugger that gets exit code, stdout/stderr from a target process.

    This class isn't actually ran as a thread, only the start_monitoring
    method is. It can spawn/stop a process, wait for it to exit and report on
    the exit status/code.
    """

    def __init__(
        self,
        start_commands,
        process_monitor,
        proc_name=None,
        ignore_pid=None,
        coredump_dir=None,
        log_level=1,
        capture_output=False,
        **kwargs
    ):
        threading.Thread.__init__(self)

        self.proc_name = proc_name
        self.ignore_pid = ignore_pid
        self.start_commands = start_commands
        self.process_monitor = process_monitor
        self.coredump_dir = coredump_dir
        self.capture_output = capture_output
        self.finished_starting = threading.Event()
        # if isinstance(start_commands, basestring):
        #     self.tokens = start_commands.split(' ')
        # else:
        #     self.tokens = start_commands
        self.cmd_args = []
        self.pid = None
        self.exit_status = None
        self.log_level = log_level
        self._process = None

    def log(self, msg="", level=1):
        """
        If the supplied message falls under the current log level, print the specified message to screen.

        @type  msg: str
        @param msg: Message to log
        """

        if self.log_level >= level:
            print("[%s] %s" % (time.strftime("%I:%M.%S"), msg))

    def spawn_target(self):
        self.log("starting target process")

        for command in self.start_commands:
            self.log("exec start command: {0}".format(command))
            try:
                if self.capture_output:
                    self._process = subprocess.Popen(command, stderr=subprocess.PIPE, stdout=subprocess.PIPE)
                else:
                    self._process = subprocess.Popen(command)
            except WindowsError as e:
                print(
                    'WindowsError {errno}: "{strerror} while starting "{cmd}"'.format(
                        errno=e.winerror, strerror=e.strerror, cmd=command
                    ),
                    file=sys.stderr,
                )
                return False
            except OSError as e:
                print(
                    'OSError {errno}: "{strerror} while starting "{cmd}"'.format(
                        errno=e.errno, strerror=e.strerror, cmd=command
                    ),
                    file=sys.stderr,
                )
                return False
        if self.proc_name:
            self.log("done. waiting for start command to terminate.")
            os.waitpid(self._process.pid, 0)
            self.log('searching for process by name "{0}"'.format(self.proc_name))
            self.watch()
            self._psutil_proc = psutil.Process(pid=self.pid)
            self.process_monitor.log("found match on pid {}".format(self.pid))
        else:
            self.log("done. target up and running, giving it 5 seconds to settle in.")
            time.sleep(5)
            self.pid = self._process.pid
        self.process_monitor.log("attached to pid: {0}".format(self.pid))

    def run(self):
        """
        self.exit_status = os.waitpid(self.pid, os.WNOHANG | os.WUNTRACED)
        while self.exit_status == (0, 0):
            self.exit_status = os.waitpid(self.pid, os.WNOHANG | os.WUNTRACED)
        """
        self.spawn_target()

        self.finished_starting.set()
        if self.proc_name:
            gone, _ = psutil.wait_procs([self._psutil_proc])
            self.exit_status = gone[0].returncode
        else:
            exit_info = os.waitpid(self.pid, 0)
            self.exit_status = exit_info[1]  # [0] is the pid

        default_reason = "Process died for unknown reason"
        if self.exit_status is not None:
            if os.WCOREDUMP(self.exit_status):
                reason = "Segmentation fault"
            elif os.WIFSTOPPED(self.exit_status):
                reason = "Stopped with signal " + str(os.WTERMSIG(self.exit_status))
            elif os.WIFSIGNALED(self.exit_status):
                reason = "Terminated with signal " + str(os.WTERMSIG(self.exit_status))
            elif os.WIFEXITED(self.exit_status):
                reason = "Exit with code - " + str(os.WEXITSTATUS(self.exit_status))
            else:
                reason = default_reason
        else:
            reason = default_reason

        outdata = None
        errdata = None
        try:
            if self._process is not None:
                outdata, errdata = self._process.communicate(timeout=POPEN_COMMUNICATE_TIMEOUT_FOR_ALREADY_DEAD_TASK)
        except subprocess.TimeoutExpired:
            self.process_monitor.log(
                msg="Expired waiting for process {0} to terminate".format(self._process.pid), level=1
            )

        msg = "[{0}] Crash. Exit code: {1}. Reason - {2}\n".format(
            time.strftime("%I:%M.%S"), self.exit_status if self.exit_status is not None else "<unknown>", reason
        )
        if errdata is not None:
            msg += "STDERR:\n{0}\n".format(errdata.decode("ascii"))
        if outdata is not None:
            msg += "STDOUT:\n{0}\n".format(outdata.decode("ascii"))
        self.process_monitor.last_synopsis = msg

    def watch(self):
        """
        Continuously loop, watching for the target process. This routine "blocks" until the target process is found.
        Update self.pid when found and return.
        """
        self.pid = None
        while not self.pid:
            for pid, name in _enumerate_processes():
                # ignore the optionally specified PID.
                if pid == self.ignore_pid:
                    continue

                if name.lower() == self.proc_name.lower():
                    self.pid = pid
                    break

    def get_exit_status(self):
        return self.exit_status

    def stop_target(self):
        try:
            os.kill(self.pid, signal.SIGKILL)
        except OSError as e:
            print(
                'Error while killing process. PID: {0} errno: {1} "{2}"'.format(self.pid, e.errno, os.strerror(e.errno))
            )
            raise e

    def pre_send(self):
        pass

    def post_send(self):
        """
        This routine is called after the fuzzer transmits a test case and returns the status of the target.

        Returns:
            bool: True if the target is still active, False otherwise.
        """
        if self.is_alive():
            return True
        else:
            with open(self.process_monitor.crash_filename, "a", encoding="utf-8") as rec_file:
                rec_file.write(self.process_monitor.last_synopsis)

            if self.process_monitor.coredump_dir is not None:
                dest = os.path.join(self.process_monitor.coredump_dir, str(self.process_monitor.test_number))
                src = _get_coredump_path()

                if src is not None:
                    self.log("moving core dump %s -> %s" % (src, dest))
                    os.rename(src, dest)
            return False



import warnings

from ..fuzz_logger_db import FuzzLoggerDbReader


class SessionInfo:
    """
    .. versionchanged:: 0.4.2
       This class has been moved into the sessions subpackage. The full path is now
       boofuzz.sessions.session_info.SessionInfo.
    """

    def __init__(self, db_filename):
        self._db_reader = FuzzLoggerDbReader(db_filename=db_filename)

    @property
    def monitor_results(self):
        return self._db_reader.failure_map

    @property
    def monitor_data(self):
        return {-1, "Monitor Data is not currently saved in the database"}

    @property
    def procmon_results(self):
        warnings.warn(
            "procmon_results has been renamed to monitor_results."
            "This alias will stop working in a future version of boofuzz",
            FutureWarning,
        )
        return self.monitor_results

    @property
    def netmon_results(self):
        warnings.warn(
            "netmon_results is now part of monitor_data" "This alias will stop working in a future version of boofuzz",
            FutureWarning,
        )
        return self.monitor_data

    @property
    def fuzz_node(self):
        return None

    @property
    def total_num_mutations(self):
        return None

    @property
    def total_mutant_index(self):
        x = next(self._db_reader.query("SELECT COUNT(*) FROM cases"))[0]
        return x

    @property
    def mutant_index(self):
        return None

    def test_case_data(self, index):
        """Return test case data object (for use by web server)

        Args:
            index (int): Test case index

        Returns:
            Test case data object
        """
        return self._db_reader.get_test_case_data(index=index)

    @property
    def is_paused(self):
        return False

    @property
    def state(self):
        return "finished"

    @property
    def exec_speed(self):
        return 0

    @property
    def runtime(self):
        return 0

    @property
    def current_test_case_name(self):
        return ""



# ASN.1 / BER TYPES (http://luca.ntop.org/Teaching/Appunti/asn1.html)

from .. import blocks, exception, helpers, primitives
from ..constants import BIG_ENDIAN


class String(blocks.Block):
    """
    [0x04][0x84][dword length][string]

    Where:

        0x04 = string
        0x84 = length is 4 bytes
    """

    def __init__(self, name, request, value, options=None):
        if not options:
            options = {}

        super(String, self).__init__(name, request)

        self.value = value
        self.options = options
        self.prefix = options.get("prefix", b"\x04")

        if not self.value:
            raise exception.SullyRuntimeError("MISSING LEGO.ber_string DEFAULT VALUE")

        str_block = blocks.Block(name + "_STR", request)
        str_block.push(primitives.String())

        self.push(blocks.Size(name + "_STR", request, endian=BIG_ENDIAN, fuzzable=True))
        self.push(str_block)

    def render(self, mutation_context=None):
        # let the parent do the initial render.
        blocks.Block.render(self)

        # TODO: What is this I don't even
        self._rendered = self.prefix + b"\x84" + self._rendered

        return helpers.str_to_bytes(self._rendered)


class Integer(blocks.Block):
    """
    [0x02][0x04][dword]

    Where:

        0x02 = integer
        0x04 = integer length is 4 bytes
    """

    def __init__(self, name, request, value, options=None):
        if not options:
            options = {}

        super(Integer, self).__init__(name, request)

        self.value = value
        self.options = options

        if not self.value:
            raise exception.SullyRuntimeError("MISSING LEGO.ber_integer DEFAULT VALUE")

        self.push(primitives.DWord(self.value, endian=BIG_ENDIAN))

    def render(self):
        # let the parent do the initial render.
        blocks.Block.render(self)

        self._rendered = b"\x02\x04" + self._rendered
        return helpers.str_to_bytes(self._rendered)



from .exception import MustImplementException


class Fuzzer:
    blocks = []

    def __init__(self):
        pass

    def send(self):
        raise MustImplementException("You must implement a send() function in your fuzzer!")

    def __repr__(self):
        return "<Fuzzer>"


class BlockBasedFuzzer(Fuzzer):
    def __init__(self):
        super(BlockBasedFuzzer, self).__init__()

    def __repr__(self):
        return "<BlockBasedFuzzer>"


class DumbFileFuzzer(Fuzzer):
    def __init__(self):
        super(DumbFileFuzzer, self).__init__()

    def __repr__(self):
        return "<FileFuzzer>"



"""
This file contains constants for the IPv4 protocol.

.. versionchanged:: 0.2.0
    ip_constants has been moved into the connections subpackage.
    The full path is now boofuzz.connections.ip_constants
"""
IPV4_PROTOCOL_UDP = 0x11
#: Theoretical maximum length of a UDP packet, based on constraints in the UDP
#: packet format.
#: WARNING! a UDP packet cannot actually be this long in the context of IPv4!
UDP_MAX_LENGTH_THEORETICAL = 65535
#: Theoretical maximum length of a UDP payload based on constraints in the
#: UDP and IPv4 packet formats.
#: WARNING! Some systems may set a payload limit smaller than this.
UDP_MAX_PAYLOAD_IPV4_THEORETICAL = 65507



import attr
import collections.abc

from .protocol_session import ProtocolSession


def mutations_list_to_dict(mutations_list_or_dict):
    if isinstance(mutations_list_or_dict, dict):
        return mutations_list_or_dict
    elif isinstance(mutations_list_or_dict, collections.abc.Iterable):
        return {mutation.qualified_name: mutation for mutation in mutations_list_or_dict}
    else:
        raise ValueError("Cannot initialize a MutationContext with mutations {0}".format(mutations_list_or_dict))


@attr.s
class MutationContext:
    """Context for current mutation(s).

    MutationContext objects are created by Session (the fuzz session manager) and passed to various Fuzzable functions
    as needed.

    For complex Fuzzable types that refer to other elements' rendered values, the implementation will typically pass
    the MutationContext along to child/referenced elements to ensure they are rendered properly.

    Note: Mutations are generated in the context of a Test Case, so a Mutation has a ProtocolSession, but a
    ProtocolSession does not necessarily have a MutationContext.
    """

    mutations = attr.ib(factory=dict, converter=mutations_list_to_dict)  # maps qualified names to a Mutation
    message_path = attr.ib(factory=list)
    protocol_session = attr.ib(type=ProtocolSession, default=None)



from ..pgraph import Edge


class Connection(Edge):
    def __init__(self, src, dst, callback=None):
        """
        Extends pgraph.edge with a callback option. This allows us to register a function to call between node
        transmissions to implement functionality such as challenge response systems. The callback method must follow
        this prototype::

            def callback(target, fuzz_data_logger, session, node, edge, *args, **kwargs)

        Where node is the node about to be sent, edge is the last edge along the current fuzz path to "node", session
        is a pointer to the session instance which is useful for snagging data such as sesson.last_recv which contains
        the data returned from the last socket transmission and sock is the live socket. A callback is also useful in
        situations where, for example, the size of the next packet is specified in the first packet.

        Args:
            src (int): Edge source ID
            dst (int): Edge destination ID
            callback (function): Optional. Callback function to pass received data to between node xmits

        .. versionchanged:: 0.4.2
           This class has been moved into the sessions subpackage. The full path is now
           boofuzz.sessions.connection.Connection.
        """

        super(Connection, self).__init__(src, dst)

        self.callback = callback



import csv
import datetime
import sys

from . import helpers, ifuzz_logger_backend


def hex_to_hexstr(input_bytes):
    """
    Render input_bytes as ASCII-encoded hex bytes, followed by a best effort
    utf-8 rendering.

    :param input_bytes: Arbitrary bytes.

    :return: Printable string.
    """
    return helpers.hex_str(input_bytes)


DEFAULT_HEX_TO_STR = hex_to_hexstr


def get_time_stamp():
    s = datetime.datetime.utcnow().isoformat()
    return s


class FuzzLoggerCsv(ifuzz_logger_backend.IFuzzLoggerBackend):
    """
    This class formats FuzzLogger data for pcap file. It can be
    configured to output to a named file.
    """

    def __init__(self, file_handle=sys.stdout, bytes_to_str=DEFAULT_HEX_TO_STR):
        """
        Args:
            file_handle (io.BinaryIO): Open file handle for logging. Defaults to sys.stdout.
            bytes_to_str (function): Function that converts sent/received bytes data to string for logging.
        """
        self._file_handle = file_handle
        self._format_raw_bytes = bytes_to_str
        self._csv_handle = csv.writer(self._file_handle)

    def open_test_step(self, description):
        self._print_log_msg(["open step", "", "", description])

    def log_check(self, description):
        self._print_log_msg(["check", "", "", description])

    def log_error(self, description):
        self._print_log_msg(["error", "", "", description])

    def log_recv(self, data):
        self._print_log_msg(["recv", len(data), self._format_raw_bytes(data), repr(data)])

    def log_send(self, data):
        self._print_log_msg(["send", len(data), self._format_raw_bytes(data), repr(data)])

    def log_info(self, description):
        self._print_log_msg(["info", "", "", description])

    def open_test_case(self, test_case_id, name, index, *args, **kwargs):
        self._print_log_msg(["open test case", "", "", "Test case " + str(test_case_id)])

    def log_fail(self, description=""):
        self._print_log_msg(["fail", "", "", description])

    def log_pass(self, description=""):
        self._print_log_msg(["pass", "", "", description])

    def close_test_case(self):
        pass

    def close_test(self):
        pass

    def _print_log_msg(self, msg):
        time_stamp = get_time_stamp()
        self._csv_handle.writerow([time_stamp] + msg)



import ssl

from boofuzz import exception
from boofuzz.connections import tcp_socket_connection


class SSLSocketConnection(tcp_socket_connection.TCPSocketConnection):
    """BaseSocketConnection implementation for use with SSL Sockets.

    .. versionadded:: 0.2.0

    Args:
        host (str): Hostname or IP adress of target system.
        port (int): Port of target service.
        send_timeout (float): Seconds to wait for send before timing out. Default 5.0.
        recv_timeout (float): Seconds to wait for recv before timing out. Default 5.0.
        server (bool): Set to True to enable server side fuzzing.
        sslcontext (ssl.SSLContext): Python SSL context to be used. Required if server=True or server_hostname=None.
        server_hostname (string): server_hostname, required for verifying identity of remote SSL/TLS server
    """

    def __init__(
        self, host, port, send_timeout=5.0, recv_timeout=5.0, server=False, sslcontext=None, server_hostname=None
    ):
        super(SSLSocketConnection, self).__init__(host, port, send_timeout, recv_timeout, server)

        self.sslcontext = sslcontext
        self.server_hostname = server_hostname

        if self.server is True and self.sslcontext is None:
            raise ValueError("Parameter sslcontext is required when server=True.")
        if self.sslcontext is None and self.server_hostname is None:
            raise ValueError("SSL/TLS requires either sslcontext or server_hostname to be set.")

    def open(self):
        # If boofuzz is the SSL client and user did not give us a SSLContext,
        # then we just use a default one.
        if self.server is False and self.sslcontext is None:
            self.sslcontext = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            self.sslcontext.check_hostname = True
            self.sslcontext.verify_mode = ssl.CERT_REQUIRED

        super(SSLSocketConnection, self)._open_socket()

        # Create SSL socket
        try:
            self._sock = self.sslcontext.wrap_socket(
                self._sock, server_side=self.server, server_hostname=self.server_hostname
            )
        except ssl.SSLError as e:
            self.close()
            raise exception.BoofuzzTargetConnectionFailedError(str(e))
        except AttributeError:
            # No SSL context set
            pass

        super(SSLSocketConnection, self)._connect_socket()

    def recv(self, max_bytes):
        """
        Receive up to max_bytes data from the target.

        Args:
            max_bytes (int): Maximum number of bytes to receive.

        Returns:
            Received data.
        """
        data = b""

        try:
            data = super(SSLSocketConnection, self).recv(max_bytes)
        except ssl.SSLError as e:
            # If an SSL error is thrown the connection should be treated as lost
            # All other exceptions should be handled / raised / re-raised by the parent class.
            raise exception.BoofuzzSSLError(str(e))

        return data

    def send(self, data):
        """
        Send data to the target. Only valid after calling open!

        Args:
            data: Data to send.

        Returns:
            int: Number of bytes actually sent.
        """
        num_sent = 0

        if len(data) > 0:
            try:
                num_sent = super(SSLSocketConnection, self).send(data)
            except ssl.SSLError as e:
                # If an SSL error is thrown the connection should be treated as lost.
                # All other exceptions should be handled / raised / re-raised by the parent class.
                raise exception.BoofuzzSSLError(str(e))

        return num_sent



import attr


@attr.s
class ProtocolSessionReference:
    """Refers to a dynamic value received or generated in the context of an individual test case.

    Pass this object as a primitive's ``default_value`` argument, and make sure you set the referred-to value using
    callbacks, e.g. ``post_test_case_callbacks`` (see :class:`Session <boofuzz.Session>`).

    Args:
        name (str): Refers to a test case session key. Must be set in the
            :class:`ProtocolSession <boofuzz.ProtocolSession>` by the time the value is required in the protocol
            definition. See :class:`Session <boofuzz.Session>`.
        default_value: The default value, used if the element must be rendered outside the context of a test
            case, or sometimes for generating mutations.
    """

    name = attr.ib(type=str)
    default_value = attr.ib()
    pass



#!/usr/bin/env python3
"""Demo FTP fuzzer as a standalone script."""

from boofuzz import *


def main():
    """
    This example is a very simple FTP fuzzer. It uses no process monitory
    (procmon) and assumes that the FTP server is already running.
    """
    session = Session(target=Target(connection=TCPSocketConnection("127.0.0.1", 21)))

    define_proto(session=session)

    session.fuzz()


def define_proto(session):
    # disable Black formatting to keep custom indentation
    # fmt: off
    user = Request("user", children=(
        String(name="key", default_value="USER"),
        Delim(name="space", default_value=" "),
        String(name="val", default_value="anonymous"),
        Static(name="end", default_value="\r\n"),
    ))

    passw = Request("pass", children=(
        String(name="key", default_value="PASS"),
        Delim(name="space", default_value=" "),
        String(name="val", default_value="james"),
        Static(name="end", default_value="\r\n"),
    ))

    stor = Request("stor", children=(
        String(name="key", default_value="STOR"),
        Delim(name="space", default_value=" "),
        String(name="val", default_value="AAAA"),
        Static(name="end", default_value="\r\n"),
    ))

    retr = Request("retr", children=(
        String(name="key", default_value="RETR"),
        Delim(name="space", default_value=" "),
        String(name="val", default_value="AAAA"),
        Static(name="end", default_value="\r\n"),
    ))
    # fmt: on

    session.connect(user)
    session.connect(user, passw)
    session.connect(passw, stor)
    session.connect(passw, retr)


def define_proto_static(session):
    """Same protocol, using the static definition style."""
    s_initialize("user")
    s_string("USER")
    s_delim(" ")
    s_string("anonymous")
    s_static("\r\n")

    s_initialize("pass")
    s_string("PASS")
    s_delim(" ")
    s_string("james")
    s_static("\r\n")

    s_initialize("stor")
    s_string("STOR")
    s_delim(" ")
    s_string("AAAA")
    s_static("\r\n")

    s_initialize("retr")
    s_string("RETR")
    s_delim(" ")
    s_string("AAAA")
    s_static("\r\n")

    session.connect(s_get("user"))
    session.connect(s_get("user"), s_get("pass"))
    session.connect(s_get("pass"), s_get("stor"))
    session.connect(s_get("pass"), s_get("retr"))


if __name__ == "__main__":
    main()



import attr

from . import helpers


@attr.s
class DataTestStep:
    type = attr.ib()
    description = attr.ib()
    data = attr.ib()
    timestamp = attr.ib()
    try:
        truncated = attr.ib(type=bool)
    except TypeError:
        # in case attr version is too old
        truncated = attr.ib()

    @property
    def text_render(self):
        return helpers.format_log_msg(
            msg_type=self.type,
            description=self.description,
            data=self.data,
            timestamp=self.timestamp,
            truncated=self.truncated,
            format_type="terminal",
        )

    @property
    def html_log_line(self):
        return helpers.format_log_msg(
            msg_type=self.type,
            description=self.description,
            data=self.data,
            timestamp=self.timestamp,
            truncated=self.truncated,
            format_type="html",
        )

    @property
    def css_class(self):
        return helpers.test_step_info[self.type]["css_class"]



import abc
import math
import os
import socket
import struct

from boofuzz.connections import itarget_connection


def _seconds_to_sockopt_format(seconds):
    """Convert floating point seconds value to second/useconds struct used by UNIX socket library.
    For Windows, convert to whole milliseconds.
    """
    if os.name == "nt":
        return int(seconds * 1000)
    else:
        microseconds_per_second = 1000000
        whole_seconds = int(math.floor(seconds))
        whole_microseconds = int(math.floor((seconds % 1) * microseconds_per_second))
        return struct.pack("ll", whole_seconds, whole_microseconds)


class BaseSocketConnection(itarget_connection.ITargetConnection, metaclass=abc.ABCMeta):
    """This class serves as a base for a number of Connections over sockets.

    .. versionadded:: 0.2.0

    Args:
        send_timeout (float): Seconds to wait for send before timing out. Default 5.0.
        recv_timeout (float): Seconds to wait for recv before timing out. Default 5.0.
    """

    def __init__(self, send_timeout, recv_timeout):
        self._send_timeout = send_timeout
        self._recv_timeout = recv_timeout

        self._sock = None

    def close(self):
        """
        Close connection to the target.

        Returns:
            None
        """
        self._sock.close()

    @abc.abstractmethod
    def open(self):
        """
        Opens connection to the target. Make sure to call close!

        Returns:
            None
        """
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_SNDTIMEO, _seconds_to_sockopt_format(self._send_timeout))
        self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVTIMEO, _seconds_to_sockopt_format(self._recv_timeout))



import ctypes
import errno
import platform
import socket
import sys

from boofuzz import exception
from boofuzz.connections import base_socket_connection, ip_constants


class UDPSocketConnection(base_socket_connection.BaseSocketConnection):
    """BaseSocketConnection implementation for use with UDP Sockets.

    .. versionadded:: 0.2.0

    Args:
        host (str): Hostname or IP adress of target system.
        port (int): Port of target service.
        send_timeout (float): Seconds to wait for send before timing out. Default 5.0.
        recv_timeout (float): Seconds to wait for recv before timing out. Default 5.0.
        server (bool): Set to True to enable server side fuzzing.
        bind (tuple (host, port)): Socket bind address and port. Required if using recv().
        broadcast (bool): Set to True to enable UDP broadcast. Must supply appropriate broadcast address for send()
            to work, and '' for bind host for recv() to work.
    """

    _max_payload = None

    def __init__(self, host, port, send_timeout=5.0, recv_timeout=5.0, server=False, bind=None, broadcast=False):
        super(UDPSocketConnection, self).__init__(send_timeout, recv_timeout)

        self.host = host
        self.port = port
        self.server = server
        self.bind = bind
        self.broadcast = broadcast

        self._serverSock = None
        self._udp_client_port = None

        self.max_payload()

        if self.bind and self.server:
            raise Exception("You cannot set both bind and server at the same time.")

    def open(self):
        """Opens connection to the target. Make sure to call close!

        Returns:
            None
        """

        self._sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        if self.bind:
            self._sock.bind(self.bind)

        if self.broadcast:
            self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, True)

        super(UDPSocketConnection, self).open()

        if self.server:
            self._sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._sock.bind((self.host, self.port))

    def recv(self, max_bytes):
        """Receive up to max_bytes data from the target.

        Args:
            max_bytes(int): Maximum number of bytes to receive.

        Returns:
            Received data.
        """
        data = b""

        try:
            if self.bind or self.server:
                data, self._udp_client_port = self._sock.recvfrom(max_bytes)
            else:
                raise exception.SullyRuntimeError(
                    "UDPSocketConnection.recv() requires a bind address/port." " Current value: {}".format(self.bind)
                )
        except socket.timeout:
            data = b""
        except socket.error as e:
            if e.errno == errno.ECONNABORTED:
                raise exception.BoofuzzTargetConnectionAborted(
                    socket_errno=e.errno, socket_errmsg=e.strerror
                ).with_traceback(sys.exc_info()[2])
            elif e.errno in [errno.ECONNRESET, errno.ENETRESET, errno.ETIMEDOUT]:
                raise exception.BoofuzzTargetConnectionReset().with_traceback(sys.exc_info()[2])
            elif e.errno == errno.EWOULDBLOCK:
                data = b""
            else:
                raise

        return data

    def send(self, data):
        """
        Send data to the target. Only valid after calling open!
        Some protocols will truncate; see self.MAX_PAYLOADS.

        Args:
            data: Data to send.

        Returns:
            int: Number of bytes actually sent.
        """
        num_sent = 0
        data = data[: self._max_payload]

        try:
            if self.server:
                if self._udp_client_port is None:
                    raise exception.BoofuzzError("recv() must be called before send with udp fuzzing servers.")

                num_sent = self._sock.sendto(data, self._udp_client_port)
            else:
                num_sent = self._sock.sendto(data, (self.host, self.port))
        except socket.error as e:
            if e.errno == errno.ECONNABORTED:
                raise exception.BoofuzzTargetConnectionAborted(
                    socket_errno=e.errno, socket_errmsg=e.strerror
                ).with_traceback(sys.exc_info()[2])
            elif e.errno in [errno.ECONNRESET, errno.ENETRESET, errno.ETIMEDOUT, errno.EPIPE]:
                raise exception.BoofuzzTargetConnectionReset().with_traceback(sys.exc_info()[2])
            else:
                raise

        return num_sent

    @classmethod
    def max_payload(cls):
        """Returns the maximum payload this connection can send at once.

        This performs some crazy CTypes magic to do a getsockopt() which determines the max UDP payload size
        in a platform-agnostic way.

        Returns:
            int: The maximum length of a UDP packet the current platform supports
        """

        # Compute max payload on first access, then re-use the cached value.
        # This code was in helpers.py, but is after the splitting of the
        # SocketConnection class more appropriate here.

        if cls._max_payload is not None:
            return cls._max_payload

        windows = platform.uname()[0] == "Windows"
        mac = platform.uname()[0] == "Darwin"
        linux = platform.uname()[0] == "Linux"
        openbsd = platform.uname()[0] == "OpenBSD"
        lib = None

        # pytype: disable=attribute-error,module-attr
        if windows:
            sol_socket = ctypes.c_int(0xFFFF)
            sol_max_msg_size = 0x2003
            lib = ctypes.WinDLL("Ws2_32.dll")
            opt = ctypes.c_int(sol_max_msg_size)
        elif linux or mac or openbsd:
            if mac:
                lib = ctypes.cdll.LoadLibrary("libc.dylib")
            elif linux:
                lib = ctypes.cdll.LoadLibrary("libc.so.6")
            elif openbsd:
                lib = ctypes.cdll.LoadLibrary("libc.so")
            sol_socket = ctypes.c_int(socket.SOL_SOCKET)
            opt = ctypes.c_int(socket.SO_SNDBUF)

        else:
            raise Exception("Unknown platform!")

        ulong_size = ctypes.sizeof(ctypes.c_ulong)
        buf = ctypes.create_string_buffer(ulong_size)
        bufsize = ctypes.c_int(ulong_size)

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

        lib.getsockopt(sock.fileno(), sol_socket, opt, buf, ctypes.pointer(bufsize))
        # pytype: enable=attribute-error,module-attr

        # Sanity filter against UDP_MAX_PAYLOAD_IPV4_THEORETICAL
        cls._max_payload = min(ctypes.c_ulong.from_buffer(buf).value, ip_constants.UDP_MAX_PAYLOAD_IPV4_THEORETICAL)

        return cls._max_payload

    @property
    def info(self):
        return "{0}:{1}".format(self.host, self.port)



# Import blocks at this level for API backwards compatibility.
from .base_primitive import BasePrimitive
from .bit_field import BitField
from .byte import Byte
from .bytes import Bytes
from .delim import Delim
from .dword import DWord
from .float import Float
from .from_file import FromFile
from .group import Group
from .mirror import Mirror
from .qword import QWord
from .random_data import RandomData
from .simple import Simple
from .static import Static
from .string import String
from .word import Word

__all__ = [
    "BasePrimitive",
    "BitField",
    "Byte",
    "Bytes",
    "Delim",
    "DWord",
    "Float",
    "FromFile",
    "Group",
    "Mirror",
    "QWord",
    "RandomData",
    "Simple",
    "Static",
    "String",
    "Word",
]



import struct

from boofuzz.primitives.bit_field import BitField


class QWord(BitField):
    """The 8 byte sized bit field primitive.

    :type  name: str, optional
    :param name: Name, for referencing later. Names should always be provided, but if not, a default name will be given,
        defaults to None
    :type  default_value: int, optional
    :param default_value: Default integer value, defaults to 0
    :type  max_num: int, optional
    :param max_num: Maximum number to iterate up to, defaults to None
    :type  endian: char, optional
    :param endian: Endianness of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >), defaults to LITTLE_ENDIAN
    :type  output_format: str, optional
    :param output_format: Output format, "binary" or "ascii", defaults to binary
    :type  signed: bool, optional
    :param signed: Make size signed vs. unsigned (applicable only with format="ascii"), defaults to False
    :type  full_range: bool, optional
    :param full_range: If enabled the field mutates through *all* possible values, defaults to False
    :type  fuzz_values: list, optional
    :param fuzz_values: List of custom fuzz values to add to the normal mutations, defaults to None
    :type  fuzzable: bool, optional
    :param fuzzable: Enable/disable fuzzing of this primitive, defaults to true
    """

    def __init__(self, *args, **kwargs):
        # Inject our width argument
        super(QWord, self).__init__(width=64, *args, **kwargs)

    def encode(self, value, mutation_context):
        if not isinstance(value, (int, list, tuple)):
            value = struct.unpack(self.endian + "Q", value)[0]
        return super(QWord, self).encode(value, mutation_context)



#!/usr/bin/env python
import logging
import shlex
import time

import click

from . import constants, sessions
from .cli_context import CliContext
from .constants import DEFAULT_PROCMON_PORT
from .connections import TCPSocketConnection
from .fuzz_logger_csv import FuzzLoggerCsv
from .fuzz_logger_curses import FuzzLoggerCurses
from .fuzz_logger_text import FuzzLoggerText
from .helpers import parse_target
from .monitors import ProcessMonitor
from .utils.process_monitor_local import ProcessMonitorLocal
from .utils.debugger_thread_simple import DebuggerThreadSimple

temp_static_session = None
temp_static_procmon = None
temp_static_fuzz_only_one_case = None


@click.group(help="boofuzz experimental CLI; usage may change over time")
def cli():
    pass


@cli.group(help="Must be run via a fuzz script")
@click.option("--target", metavar="HOST:PORT", help="Target network address", required=True)
@click.option("--test-case-index", help="Test case index", type=str)
@click.option("--test-case-name", help="Name of node or specific test case")
@click.option("--csv-out", help="Output to CSV file")
@click.option(
    "--sleep-between-cases", help="Wait FLOAT (seconds) between test cases (partial seconds OK)", type=float, default=0
)
@click.option("--procmon-host", help="Process monitor port host or IP")
@click.option("--procmon-port", type=int, default=DEFAULT_PROCMON_PORT, help="Process monitor port")
@click.option("--procmon-start", help="Process monitor start command")
@click.option("--procmon-capture", is_flag=True, help="Capture stdout/stderr from target process upon failure")
@click.option("--tui/--no-tui", help="Enable/disable TUI")
@click.option("--text-dump/--no-text-dump", help="Enable/disable full text dump of logs", default=False)
@click.option("--feature-check", is_flag=True, help="Run a feature check instead of a fuzz test", default=False)
@click.option("--target-cmd", help="Target command and arguments")
@click.option(
    "--keep-web/--no-keep-web",
    is_flag=True,
    default=True,
    help="Keep web server for web UI open when out of fuzz cases",
)
@click.option(
    "--combinatorial/--no-combinatorial", is_flag=True, default=True, help="Enable fuzzing with multiple mutations"
)
@click.option(
    "--record-passes",
    default=10,
    type=int,
    help="Record this many cases before each failure. Set to 0 to record all test cases (high disk space usage!).",
)
@click.pass_context
def fuzz(
    ctx,
    target,
    test_case_index,
    test_case_name,
    csv_out,
    sleep_between_cases,
    procmon_host,
    procmon_port,
    procmon_start,
    procmon_capture,
    tui,
    text_dump,
    feature_check,
    target_cmd,
    keep_web,
    combinatorial,
    record_passes,
):
    local_procmon = None
    if target_cmd is not None and procmon_host is None:
        local_procmon = ProcessMonitorLocal(
            crash_filename="boofuzz-crash-bin",
            proc_name=None,
            pid_to_ignore=None,
            debugger_class=DebuggerThreadSimple,
            level=1,
        )

    fuzz_loggers = []
    if text_dump:
        fuzz_loggers.append(FuzzLoggerText())
    elif tui:
        fuzz_loggers.append(FuzzLoggerCurses())
    if csv_out is not None:
        f = open("boofuzz.csv", "wb")
        fuzz_loggers.append(FuzzLoggerCsv(file_handle=f))

    procmon_options = {}
    if procmon_start is not None:
        procmon_options["start_commands"] = [procmon_start]
    if target_cmd is not None:
        procmon_options["start_commands"] = shlex.split(target_cmd)
    if procmon_capture:
        procmon_options["capture_output"] = True

    if local_procmon is not None or procmon_host is not None:
        if procmon_host is not None:
            procmon = ProcessMonitor(procmon_host, procmon_port)
        else:
            procmon = local_procmon
        procmon.set_options(**procmon_options)
        monitors = [procmon]
    else:
        procmon = None
        monitors = []

    if combinatorial:
        max_depth = None
    else:
        max_depth = 1

    if test_case_index is None:
        start = 1
        end = None
    elif "-" in test_case_index:
        start, end = test_case_index.split("-")
        if not start:
            start = 1
        else:
            start = int(start)
        if not end:
            end = None
        else:
            end = int(end)
    else:
        start = end = int(test_case_index)

    connection = TCPSocketConnection(*parse_target(target_name=target))

    session = sessions.Session(
        target=sessions.Target(
            connection=connection,
            monitors=monitors,
        ),
        fuzz_loggers=fuzz_loggers,
        sleep_time=sleep_between_cases,
        index_start=start,
        index_end=end,
        keep_web_open=keep_web,
        fuzz_db_keep_only_n_pass_cases=record_passes,
    )

    ctx.obj = CliContext(session=session)

    # The resultcallback is called after any subcommands, e.g. the one provided by the user
    @fuzz.resultcallback()
    def fuzzcallback(result, *args, **kwargs):
        if feature_check:
            session.feature_check()
        else:
            session.fuzz(name=test_case_name, max_depth=max_depth)

        if procmon is not None:
            procmon.stop_target()


@cli.command(name="open")
@click.option("--debug", help="Print debug info to console", is_flag=True)
@click.option(
    "--ui-port",
    help="Port on which to serve the web interface (default {0})".format(constants.DEFAULT_PROCMON_PORT),
    type=int,
    default=constants.DEFAULT_WEB_UI_PORT,
)
@click.option(
    "--ui-addr",
    help="Address on which to serve the web interface (default localhost). Set to empty "
    "string to serve on all interfaces.",
    type=str,
    default="localhost",
)
@click.argument("filename")
def open_file(debug, filename, ui_port, ui_addr):
    if debug:
        logging.basicConfig(level=logging.DEBUG)

    sessions.open_test_run(db_filename=filename, port=ui_port, address=ui_addr)

    print("Serving web page at http://{0}:{1}. Hit Ctrl+C to quit.".format(ui_addr, ui_port))
    while True:
        time.sleep(0.001)


def main():
    cli()


def main_helper(click_command=None):
    """
    Args:
        click_command (click.Command): Click command to add as a sub-command to boo fuzz.
    """
    if click_command is not None:
        fuzz.add_command(click_command)
    main()


if __name__ == "__main__":
    main()



import math
import struct

from .. import helpers


def bind(uuid, version):
    """
    Generate the data necessary to bind to the specified interface.
    """

    major, minor = version.split(".")

    major = struct.pack("<H", int(major))
    minor = struct.pack("<H", int(minor))

    bind_request = b"\x05\x00"  # version 5.0
    bind_request += b"\x0b"  # packet type = bind (11)
    bind_request += b"\x03"  # packet flags = last/first flag set
    bind_request += b"\x10\x00\x00\x00"  # data representation
    bind_request += b"\x48\x00"  # frag length: 72
    bind_request += b"\x00\x00"  # auth length
    bind_request += b"\x00\x00\x00\x00"  # call id
    bind_request += b"\xb8\x10"  # max transmit frag (4280)
    bind_request += b"\xb8\x10"  # max recv frag (4280)
    bind_request += b"\x00\x00\x00\x00"  # assoc group
    bind_request += b"\x01"  # number of ctx items (1)
    bind_request += b"\x00\x00\x00"  # padding
    bind_request += b"\x00\x00"  # context id (0)
    bind_request += b"\x01"  # number of trans items (1)
    bind_request += b"\x00"  # padding
    bind_request += helpers.uuid_str_to_bin(uuid)  # abstract syntax
    bind_request += major  # interface version
    bind_request += minor  # interface version minor

    # transfer syntax 8a885d04-1ceb-11c9-9fe8-08002b104860 v2.0
    bind_request += b"\x04\x5d\x88\x8a\xeb\x1c\xc9\x11\x9f\xe8\x08\x00\x2b\x10\x48\x60"
    bind_request += b"\x02\x00\x00\x00"

    return bind_request


def bind_ack(data):
    """
    Ensure the data is a bind ack and that the
    """

    # packet type == bind ack (12)?
    if data[2] != b"\x0c":
        return False

    # ack result == acceptance?
    if data[36:38] != b"\x00\x00":
        return False

    return True


def request(opnum, data):
    """
    Return a list of packets broken into 5k fragmented chunks necessary to make the RPC request.
    """

    frag_size = 1000  # max frag size = 5840?
    frags = []

    num_frags = int(math.ceil(float(len(data)) / float(frag_size)))

    for i in range(num_frags):
        chunk = data[i * frag_size : (i + 1) * frag_size]
        frag_length = struct.pack("<H", len(chunk) + 24)
        alloc_hint = struct.pack("<L", len(chunk))

        flags = 0
        if i == 0:
            flags |= 0x1  # first frag
        if i == num_frags - 1:
            flags |= 0x2  # last frag

        request_buffer = b"\x05\x00"  # version 5.0
        request_buffer += b"\x00"  # packet type = request (0)
        request_buffer += struct.pack("B", flags)  # packet flags
        request_buffer += b"\x10\x00\x00\x00"  # data representation
        request_buffer += frag_length  # frag length
        request_buffer += b"\x00\x00"  # auth length
        request_buffer += b"\x00\x00\x00\x00"  # call id
        request_buffer += alloc_hint  # alloc hint
        request_buffer += b"\x00\x00"  # context id (0)
        request_buffer += struct.pack("<H", opnum)  # opnum
        request_buffer += chunk

        frags.append(request_buffer)

    # you don't have to send chunks out individually. so make life easier for the user and send them all at once.
    return b"".join(frags)



# Import connections at this level for API backwards compatibility.
from .base_socket_connection import BaseSocketConnection
from .file_connection import FileConnection
from .iserial_like import ISerialLike
from .itarget_connection import ITargetConnection
from .netconf_connection import NETCONFConnection
from .raw_l2_socket_connection import RawL2SocketConnection
from .raw_l3_socket_connection import RawL3SocketConnection
from .serial_connection import SerialConnection
from .serial_connection_low_level import SerialConnectionLowLevel
from .socket_connection import SocketConnection
from .ssl_socket_connection import SSLSocketConnection
from .tcp_socket_connection import TCPSocketConnection
from .udp_socket_connection import UDPSocketConnection
from .unix_socket_connection import UnixSocketConnection

__all__ = [
    "BaseSocketConnection",
    "FileConnection",
    "ISerialLike",
    "ITargetConnection",
    "NETCONFConnection",
    "RawL2SocketConnection",
    "RawL3SocketConnection",
    "SerialConnection",
    "SerialConnectionLowLevel",
    "SocketConnection",
    "SSLSocketConnection",
    "TCPSocketConnection",
    "UDPSocketConnection",
    "UnixSocketConnection",
]



from typing import Union  # noqa: F401

from .ifuzz_logger import IFuzzLogger


class FuzzLogger(IFuzzLogger):
    """
    Takes a list of IFuzzLogger objects and multiplexes logged data to each one.

    FuzzLogger also maintains summary failure and error data.

    Args:
        fuzz_loggers (:obj:`list` of :obj:`IFuzzLogger`): IFuzzLogger objects
                                                          to which to send log data.
    """

    def __init__(self, fuzz_loggers=None):
        if fuzz_loggers is None:
            fuzz_loggers = []
        self._fuzz_loggers = fuzz_loggers

        self._cur_test_case_id = ""  # type: Union[int, str]
        self.failed_test_cases = {}
        self.error_test_cases = {}
        self.passed_test_case_count = 0  # passed test cases are simply counted to avoid runaway memory usage
        self._last_passed_id = ""  # helps avoid duplicates
        self.test_case_count = 0

    @property
    def most_recent_test_id(self):
        """Return a value (e.g. string) representing the most recent test case."""
        return self._cur_test_case_id

    def open_test_step(self, description):
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.open_test_step(description=description)

    def log_error(self, description):
        if self._cur_test_case_id not in self.error_test_cases:
            self.error_test_cases[self._cur_test_case_id] = []
        self.error_test_cases[self._cur_test_case_id].append(description)
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.log_error(description=description)

    def log_fail(self, description=""):
        if self._cur_test_case_id not in self.failed_test_cases:
            self.failed_test_cases[self._cur_test_case_id] = []
        self.failed_test_cases[self._cur_test_case_id].append(description)
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.log_fail(description=description)

    def log_info(self, description):
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.log_info(description=description)

    def log_recv(self, data):
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.log_recv(data=data)

    def log_pass(self, description=""):
        if self._cur_test_case_id != self._last_passed_id:
            self.passed_test_case_count += 1
            self._last_passed_id = self._cur_test_case_id
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.log_pass(description=description)

    def log_check(self, description):
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.log_check(description=description)

    def open_test_case(self, test_case_id, name, index, *args, **kwargs):
        self._cur_test_case_id = test_case_id
        self.test_case_count += 1
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.open_test_case(test_case_id=test_case_id, name=name, index=index, *args, **kwargs)

    def log_send(self, data):
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.log_send(data=data)

    def close_test_case(self):
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.close_test_case()

    def close_test(self):
        for fuzz_logger in self._fuzz_loggers:
            fuzz_logger.close_test()

    def failure_summary(self):
        """Return test summary string based on fuzz logger results.

        :return: Test summary string, may be multi-line.
        """
        summary = "Test Summary: {0} tests ran.\n".format(self.test_case_count)
        summary += "PASSED: {0} test cases.\n".format(self.passed_test_case_count)

        if len(self.failed_test_cases) > 0:
            summary += "FAILED: {0} test cases:\n".format(len(self.failed_test_cases))
            summary += "{0}\n".format("\n".join(map(str, self.failed_test_cases)))

        if len(self.error_test_cases) > 0:
            summary += "Errors on {0} test cases:\n".format(len(self.error_test_cases))
            summary += "{0}".format("\n".join(map(str, self.error_test_cases)))

        return summary



#!/usr/bin/env python3
# Designed for use with boofuzz v0.2.0

# More advanced request definitions can be found in the request_definitions directory.

from boofuzz import *


def main():
    session = Session(
        target=Target(connection=TCPSocketConnection("127.0.0.1", 80)),
    )

    s_initialize(name="Request")
    with s_block("Request-Line"):
        s_group("Method", ["GET", "HEAD", "POST", "PUT", "DELETE", "CONNECT", "OPTIONS", "TRACE"])
        s_delim(" ", name="space-1")
        s_string("/index.html", name="Request-URI")
        s_delim(" ", name="space-2")
        s_string("HTTP/1.1", name="HTTP-Version")
        s_static("\r\n", name="Request-Line-CRLF")
        s_string("Host:", name="Host-Line")
        s_delim(" ", name="space-3")
        s_string("example.com", name="Host-Line-Value")
        s_static("\r\n", name="Host-Line-CRLF")
        s_static("Content-Length:", name="Content-Length-Header")
        s_delim(" ", name="space-4")
        s_size("Body-Content", output_format="ascii", name="Content-Length-Value")
        s_static("\r\n", "Content-Length-CRLF")
    s_static("\r\n", "Request-CRLF")

    with s_block("Body-Content"):
        s_string("Body content ...", name="Body-Content-Value")

    session.connect(s_get("Request"))

    session.fuzz()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Designed for use with boofuzz v0.2.0
#
# A partial MDNS fuzzer. Could be made to be a DNS fuzzer trivially
# Charlie Miller <cmiller@securityevaluators.com>

from boofuzz import *


def insert_questions(target, fuzz_data_logger, session, node, edge, *args, **kwargs):
    # print(node.names)
    node.names["query.Questions"].value = 1 + node.names["query.queries"].current_reps
    node.names["query.Authority"].value = 1 + node.names["query.auth_nameservers"].current_reps


s_initialize("query")
s_word(0, name="TransactionID")
s_word(0, name="Flags")
s_word(1, name="Questions", endian=">")
s_word(0, name="Answer", endian=">")
s_word(1, name="Authority", endian=">")
s_word(0, name="Additional", endian=">")

# ######## Queries ################
if s_block_start("query"):
    if s_block_start("name_chunk"):
        s_size("string", length=1)
        if s_block_start("string"):
            s_string("A" * 10)
        s_block_end()
    s_block_end()
    s_repeat("name_chunk", min_reps=2, max_reps=4, step=1, fuzzable=True, name="aName")

    s_group("end", values=[b"\x00", b"\xc0\xb0"])  # very limited pointer fuzzing
    s_word(0xC, name="Type", endian=">")
    s_word(0x8001, name="Class", endian=">")
s_block_end()
s_repeat("query", 0, 1000, 40, name="queries")


# ####### Authorities #############
if s_block_start("auth_nameserver"):
    if s_block_start("name_chunk_auth"):
        s_size("string_auth", length=1)
        if s_block_start("string_auth"):
            s_string("A" * 10)
        s_block_end()
    s_block_end()
    s_repeat("name_chunk_auth", min_reps=2, max_reps=4, step=1, fuzzable=True, name="aName_auth")
    s_group("end_auth", values=[b"\x00", b"\xc0\xb0"])  # very limited pointer fuzzing

    s_word(0xC, name="Type_auth", endian=">")
    s_word(0x8001, name="Class_auth", endian=">")
    s_dword(0x78, name="TTL_auth", endian=">")
    s_size("data_length", length=2, endian=">")
    if s_block_start("data_length"):
        s_binary("00 00 00 00 00 16 c0 b0")  # This should be fuzzed according to the type, but I'm too lazy atm
    s_block_end()
s_block_end()
s_repeat("auth_nameserver", 0, 1000, 40, name="auth_nameservers")

s_word(0)

sess = Session(target=Target(connection=UDPSocketConnection("224.0.0.251", 5353)))
sess.connect(s_get("query"), callback=insert_questions)

sess.fuzz()
